//  !!!! Modified for t3m by MC
//
//  SnapPeaC.c
//
//  The Python interface to the SnapPea kernel consists of two parts.
//
//      SnapPea.py (in a different file)(this file) defines a set
//          of objects (Triangulation, AbelianGroup, etc.) purely in Python.
//
//      SnapPeaC.c (this file) implements SnapPea.py's methods as a set
//          of wrappers to the standard SnapPea kernel functions,
//          which are written in C.
//
//  Technical comment:  There's the awkward question of how the Python
//  interpreter is to keep copies of pointers to SnapPea data structures,
//  such as Triangulations, which SnapPea.h exports as "opaque typedefs".
//  The present solution is to pass them to the interpreter as long ints.

//  If your compiler can't find Python.h, please go to the makefile
//  and adjust the line
// 
//      PYTHON_DIRECTORY = /usr/include/python1.5
//
//  If you don't know where Python.h is, try "locate Python.h".
#include <Python.h>

#include <ctype.h>
#include "assert.h"
#include "kernel.h"

//  from SnapPea's unix kit
#include "unix_cusped_census.h"
#include "unix_file_io.h"

//  implementation of the UI functions for python

// SnapPea's exceptions
// PyObject *SnapPea_Ack = NULL;
PyObject *SnapPea_Query = NULL;
PyObject *SnapPea_Fatal = NULL;

#include "SnapPea.h"
#include <stdio.h>
#include <stdlib.h>

void uAcknowledge(
    const char *message)
{
  PyErr_Warn(PyExc_UserWarning, (char *)message);
}

void uFatalError(
    char    *function,
    char    *file)
{
    char message[256];
    snprintf(message, 256,
        "A fatal error has occurred in the function %s() in the file %s.c.\n",
        function,
        file);
    PyErr_SetString(SnapPea_Fatal, message);
}

void uAbortMemoryFull(void)
{
    PyErr_NoMemory();
}

int uQuery(
    const char  *message,
    const int   num_responses,
    const char  *responses[],
    const int   default_response)
{
    /*
     *  If desired you could write this function to obtain a response from the user,
     *  but for now it set up to return the default response, to facilitate batch
     *  computations.
     */
    char error_message[256];
    snprintf(error_message, 256, "Q: %s\nA:  %s\n", message, responses[default_response]);
    PyErr_SetString(SnapPea_Query, error_message);
    return default_response;
}

/*
 *  The "long computation" feature is unused, but we define its
 *  global variables and functions to avoid a link error.
 */
Boolean gLongComputationInProgress,
        gLongComputationCancelled;
void uLongComputationBegins(
    char    *message,
    Boolean is_abortable)
{
}
FuncResult uLongComputationContinues()
{
    return func_OK;
}
void uLongComputationEnds()
{
}


// declare this here for now

static PyObject* int_array_to_PyList(int* array, int length);
Triangulation* easy_triangulate_punctured_torus_bundle(
               Boolean negative_determinant, 
	       Boolean negative_trace, char* LR_factors);

extern Triangulation* sendable_to_triangulation(char* sendable);
extern void set_tet_shapes(Triangulation *theTriangulation, Complex* theShapes);

#define FALSE   0
#define TRUE    1

void initSnapPeaC(void);
static char *RelationToString(int *aSnapPeaRelation);

#define DECLARE_WRAPPER(name)   \
    static PyObject *name(PyObject *self, PyObject *args);

DECLARE_WRAPPER(wrap_verify_my_malloc_usage);
DECLARE_WRAPPER(wrap_get_triangulation);
DECLARE_WRAPPER(wrap_save_triangulation);
DECLARE_WRAPPER(wrap_copy_triangulation);
DECLARE_WRAPPER(wrap_get_cusped_census_manifold);
DECLARE_WRAPPER(wrap_free_triangulation);
DECLARE_WRAPPER(wrap_get_triangulation_name);
DECLARE_WRAPPER(wrap_set_triangulation_name);
DECLARE_WRAPPER(wrap_get_triangulation_is_orientable);
DECLARE_WRAPPER(wrap_get_solution_type);
DECLARE_WRAPPER(wrap_get_num_tetrahedra);
DECLARE_WRAPPER(wrap_get_num_cusps);
DECLARE_WRAPPER(wrap_get_cusp_is_complete);
DECLARE_WRAPPER(wrap_get_cusp_is_orientable);
DECLARE_WRAPPER(wrap_get_cusp_m);
DECLARE_WRAPPER(wrap_get_cusp_l);
DECLARE_WRAPPER(wrap_get_holonomies);
DECLARE_WRAPPER(wrap_get_angle_sums);
DECLARE_WRAPPER(wrap_set_cusp_info);
DECLARE_WRAPPER(wrap_cusp_is_fillable);
DECLARE_WRAPPER(wrap_remove_Dehn_fillings);
DECLARE_WRAPPER(wrap_volume);
DECLARE_WRAPPER(wrap_homology);
DECLARE_WRAPPER(wrap_fill_cusp);
DECLARE_WRAPPER(wrap_make_finite_triangulation);
DECLARE_WRAPPER(wrap_get_drillable_curves);
DECLARE_WRAPPER(wrap_drill_curve);
DECLARE_WRAPPER(wrap_get_normal_surfaces);
DECLARE_WRAPPER(wrap_split_along_normal_surface);
DECLARE_WRAPPER(wrap_fundamental_group);
DECLARE_WRAPPER(wrap_free_group_presentation);
DECLARE_WRAPPER(wrap_fg_get_num_generators);
DECLARE_WRAPPER(wrap_fg_get_relations);
DECLARE_WRAPPER(wrap_fg_representation);
DECLARE_WRAPPER(wrap_fg_peripheral_curves);
DECLARE_WRAPPER(wrap_core_geodesic);
DECLARE_WRAPPER(wrap_change_peripheral_curves);
DECLARE_WRAPPER(wrap_shortest_curves_become_meridians);
DECLARE_WRAPPER(wrap_current_fillings_become_meridians);
DECLARE_WRAPPER(wrap_basic_simplification);
DECLARE_WRAPPER(wrap_randomize_triangulation);
DECLARE_WRAPPER(wrap_reorient);
DECLARE_WRAPPER(wrap_proto_canonize);
DECLARE_WRAPPER(wrap_is_canonical_triangulation);
DECLARE_WRAPPER(wrap_are_isometric);
DECLARE_WRAPPER(wrap_symmetry_group);
DECLARE_WRAPPER(wrap_free_symmetry_group);
DECLARE_WRAPPER(wrap_symmetry_group_order);
DECLARE_WRAPPER(wrap_symmetry_group_is_abelian);
DECLARE_WRAPPER(wrap_symmetry_group_abelian_description);
DECLARE_WRAPPER(wrap_symmetry_group_is_dihedral);
DECLARE_WRAPPER(wrap_symmetry_group_is_polyhedral);
DECLARE_WRAPPER(wrap_symmetry_group_polyhedral_description);
DECLARE_WRAPPER(wrap_symmetry_group_is_S5);
DECLARE_WRAPPER(wrap_symmetry_group_is_direct_product);
DECLARE_WRAPPER(wrap_symmetry_group_factor);
DECLARE_WRAPPER(wrap_symmetry_group_is_amphicheiral);
DECLARE_WRAPPER(wrap_symmetry_group_invertible_knot);
DECLARE_WRAPPER(wrap_symmetry_group_commutator_subgroup);
DECLARE_WRAPPER(wrap_symmetry_group_abelianization);
DECLARE_WRAPPER(wrap_symmetry_group_center);
DECLARE_WRAPPER(wrap_symmetry_group_presentation);
DECLARE_WRAPPER(wrap_tet_shapes);
DECLARE_WRAPPER(wrap_set_tet_shapes);
DECLARE_WRAPPER(wrap_set_target_holonomy);
DECLARE_WRAPPER(wrap_Dirichlet);
DECLARE_WRAPPER(wrap_free_Dirichlet_domain);
DECLARE_WRAPPER(wrap_Dirichlet_num_vertices);
DECLARE_WRAPPER(wrap_Dirichlet_num_edges);
DECLARE_WRAPPER(wrap_Dirichlet_num_faces);
DECLARE_WRAPPER(wrap_Dirichlet_vertices);
DECLARE_WRAPPER(wrap_Dirichlet_faces);
DECLARE_WRAPPER(wrap_Dirichlet_face_colors);
DECLARE_WRAPPER(wrap_Dirichlet_face_pairings);
DECLARE_WRAPPER(wrap_get_current_cusp_shape);
DECLARE_WRAPPER(wrap_get_current_cusp_modulus);
DECLARE_WRAPPER(get_gluing_data);
DECLARE_WRAPPER(get_triangulation_from_DT);
DECLARE_WRAPPER(get_triangulation_from_PythonKLP);
DECLARE_WRAPPER(wrap_easy_triangulate_punctured_torus_bundle);
DECLARE_WRAPPER(get_gluing_equations);
DECLARE_WRAPPER(get_cusp_equation);
DECLARE_WRAPPER(wrap_same_triangulation);
DECLARE_WRAPPER(wrap_install_shortest_bases);
DECLARE_WRAPPER(wrap_install_current_curve_bases);
DECLARE_WRAPPER(wrap_sendable_to_triangulation);
DECLARE_WRAPPER(wrap_hack_double_cover);
DECLARE_WRAPPER(wrap_hack_create_all_covers);
DECLARE_WRAPPER(wrap_choose_generators);
DECLARE_WRAPPER(wrap_hack_hide_hyperbolic_structure);
DECLARE_WRAPPER(wrap_find_complete_hyperbolic_structure);
DECLARE_WRAPPER(wrap_triangulate_link_complement_from_file);
DECLARE_WRAPPER(wrap_fibered_manifold_associated_to_braid);
DECLARE_WRAPPER(wrap_create_cover);
DECLARE_WRAPPER(wrap_get_cusp_neighborhood_info);
DECLARE_WRAPPER(get_peripheral_curve_datum);

static struct PyMethodDef SnapPeaCMethods[] =
{
    {"verify_my_malloc_usage",                  wrap_verify_my_malloc_usage,                METH_VARARGS},
    {"get_triangulation",                       wrap_get_triangulation,                     METH_VARARGS},
    {"save_triangulation",                      wrap_save_triangulation,                    METH_VARARGS},
    {"copy_triangulation",                      wrap_copy_triangulation,                    METH_VARARGS},
    {"get_cusped_census_manifold",              wrap_get_cusped_census_manifold,            METH_VARARGS},
    {"free_triangulation",                      wrap_free_triangulation,                    METH_VARARGS},
    {"get_triangulation_name",                  wrap_get_triangulation_name,                METH_VARARGS},
    {"set_triangulation_name",                  wrap_set_triangulation_name,                METH_VARARGS},
    {"get_triangulation_is_orientable",         wrap_get_triangulation_is_orientable,       METH_VARARGS},
    {"get_solution_type",                       wrap_get_solution_type,                     METH_VARARGS},
    {"get_num_tetrahedra",                      wrap_get_num_tetrahedra,                    METH_VARARGS},
    {"get_num_cusps",                           wrap_get_num_cusps,                         METH_VARARGS},
    {"get_cusp_is_complete",                    wrap_get_cusp_is_complete,                  METH_VARARGS},
    {"get_cusp_is_orientable",                  wrap_get_cusp_is_orientable,                METH_VARARGS},
    {"get_cusp_m",                              wrap_get_cusp_m,                            METH_VARARGS},
    {"get_cusp_l",                              wrap_get_cusp_l,                            METH_VARARGS},
    {"get_holonomies",                          wrap_get_holonomies,                        METH_VARARGS},
    {"get_angle_sums",                          wrap_get_angle_sums,                        METH_VARARGS},
    {"set_cusp_info",                           wrap_set_cusp_info,                         METH_VARARGS},
    {"cusp_is_fillable",                        wrap_cusp_is_fillable,                      METH_VARARGS},
    {"remove_Dehn_fillings",                    wrap_remove_Dehn_fillings,                  METH_VARARGS},
    {"volume",                                  wrap_volume,                                METH_VARARGS},
    {"homology",                                wrap_homology,                              METH_VARARGS},
    {"fill_cusp",                               wrap_fill_cusp,                             METH_VARARGS},
    {"make_finite_triangulation",                            wrap_make_finite_triangulation,                             METH_VARARGS},
    {"get_drillable_curves",                    wrap_get_drillable_curves,                  METH_VARARGS},
    {"drill_curve",                             wrap_drill_curve,                           METH_VARARGS},
    {"get_normal_surfaces",                     wrap_get_normal_surfaces,                   METH_VARARGS},
    {"split_along_normal_surface",              wrap_split_along_normal_surface,            METH_VARARGS},
    {"fundamental_group",                       wrap_fundamental_group,                     METH_VARARGS},
    {"free_group_presentation",                 wrap_free_group_presentation,               METH_VARARGS},
    {"fg_get_num_generators",                   wrap_fg_get_num_generators,                 METH_VARARGS},
    {"fg_get_relations",                        wrap_fg_get_relations,                      METH_VARARGS},
    {"fg_representation",                       wrap_fg_representation,                     METH_VARARGS},
    {"fg_peripheral_curves",                    wrap_fg_peripheral_curves,                  METH_VARARGS},
    {"core_geodesic",                           wrap_core_geodesic,                         METH_VARARGS},
    {"change_peripheral_curves",                wrap_change_peripheral_curves,              METH_VARARGS},
    {"shortest_curves_become_meridians",        wrap_shortest_curves_become_meridians,      METH_VARARGS},
    {"current_fillings_become_meridians",       wrap_current_fillings_become_meridians,     METH_VARARGS},
    {"basic_simplification",                    wrap_basic_simplification,                  METH_VARARGS},
    {"randomize_triangulation",                 wrap_randomize_triangulation,               METH_VARARGS},
    {"reorient",                                wrap_reorient,                              METH_VARARGS},
    {"proto_canonize",                          wrap_proto_canonize,                        METH_VARARGS},
    {"is_canonical_triangulation",              wrap_is_canonical_triangulation,            METH_VARARGS},
    {"are_isometric",                           wrap_are_isometric,                         METH_VARARGS},
    {"symmetry_group",                          wrap_symmetry_group,                        METH_VARARGS},
    {"free_symmetry_group",                     wrap_free_symmetry_group,                   METH_VARARGS},
    {"symmetry_group_order",                    wrap_symmetry_group_order,                  METH_VARARGS},
    {"symmetry_group_is_abelian",               wrap_symmetry_group_is_abelian,             METH_VARARGS},
    {"symmetry_group_abelian_description",      wrap_symmetry_group_abelian_description,    METH_VARARGS},
    {"symmetry_group_is_dihedral",              wrap_symmetry_group_is_dihedral,            METH_VARARGS},
    {"symmetry_group_is_polyhedral",            wrap_symmetry_group_is_polyhedral,          METH_VARARGS},
    {"symmetry_group_polyhedral_description",   wrap_symmetry_group_polyhedral_description, METH_VARARGS},
    {"symmetry_group_is_S5",                    wrap_symmetry_group_is_S5,                  METH_VARARGS},
    {"symmetry_group_is_direct_product",        wrap_symmetry_group_is_direct_product,      METH_VARARGS},
    {"symmetry_group_factor",                   wrap_symmetry_group_factor,                 METH_VARARGS},
    {"symmetry_group_is_amphicheiral",          wrap_symmetry_group_is_amphicheiral,        METH_VARARGS},
    {"symmetry_group_invertible_knot",          wrap_symmetry_group_invertible_knot,        METH_VARARGS},
    {"symmetry_group_commutator_subgroup",      wrap_symmetry_group_commutator_subgroup,    METH_VARARGS},
    {"symmetry_group_abelianization",           wrap_symmetry_group_abelianization,         METH_VARARGS},
    {"symmetry_group_center",                   wrap_symmetry_group_center,                 METH_VARARGS},
    {"symmetry_group_presentation",             wrap_symmetry_group_presentation,           METH_VARARGS},
    {"tet_shapes",                              wrap_tet_shapes,                            METH_VARARGS},
    {"set_tet_shapes",                          wrap_set_tet_shapes,                        METH_VARARGS},
    {"set_target_holonomy",                     wrap_set_target_holonomy,                   METH_VARARGS},
    {"Dirichlet",                               wrap_Dirichlet,                             METH_VARARGS},
    {"free_Dirichlet_domain",                   wrap_free_Dirichlet_domain,                 METH_VARARGS},
    {"Dirichlet_num_vertices",                  wrap_Dirichlet_num_vertices,                METH_VARARGS},
    {"Dirichlet_num_edges",                     wrap_Dirichlet_num_edges,                   METH_VARARGS},
    {"Dirichlet_num_faces",                     wrap_Dirichlet_num_faces,                   METH_VARARGS},
    {"Dirichlet_vertices",                      wrap_Dirichlet_vertices,                    METH_VARARGS},
    {"Dirichlet_faces",                         wrap_Dirichlet_faces,                       METH_VARARGS},
    {"Dirichlet_face_colors",                   wrap_Dirichlet_face_colors,                 METH_VARARGS},
    {"Dirichlet_face_pairings",                 wrap_Dirichlet_face_pairings,               METH_VARARGS},
    {"get_current_cusp_shape",                  wrap_get_current_cusp_shape,                METH_VARARGS},
    {"get_current_cusp_modulus",                wrap_get_current_cusp_modulus,              METH_VARARGS},
    {"get_gluing_data",                         get_gluing_data,                            METH_VARARGS},
    {"get_triangulation_from_DT",               get_triangulation_from_DT,                  METH_VARARGS},
    {"get_triangulation_from_PythonKLP",        get_triangulation_from_PythonKLP,           METH_VARARGS},
    {"easy_triangulate_punctured_torus_bundle", wrap_easy_triangulate_punctured_torus_bundle,METH_VARARGS},
    {"get_gluing_equations",                    get_gluing_equations,                       METH_VARARGS},
    {"get_cusp_equation",                       get_cusp_equation,                          METH_VARARGS},
    {"same_triangulation",                      wrap_same_triangulation,                    METH_VARARGS},
    {"install_shortest_bases",                  wrap_install_shortest_bases,                METH_VARARGS},
    {"install_current_curve_bases",             wrap_install_current_curve_bases,           METH_VARARGS},
    {"sendable_to_triangulation",               wrap_sendable_to_triangulation,             METH_VARARGS},
    {"hack_double_cover",                       wrap_hack_double_cover,                     METH_VARARGS},
    {"hack_create_all_covers",                       wrap_hack_create_all_covers,                     METH_VARARGS},
    {"choose_generators",                       wrap_choose_generators,                     METH_VARARGS}, 
    {"hack_hide_hyperbolic_structure",          wrap_hack_hide_hyperbolic_structure,        METH_VARARGS}, 
    {"find_complete_hyperbolic_structure",      wrap_find_complete_hyperbolic_structure,    METH_VARARGS}, 
    {"triangulate_link_complement_from_file", wrap_triangulate_link_complement_from_file, METH_VARARGS},
    {"fibered_manifold_associated_to_braid", wrap_fibered_manifold_associated_to_braid, METH_VARARGS},
 {"create_cover", wrap_create_cover, METH_VARARGS},
{"get_cusp_neighborhood_info", wrap_get_cusp_neighborhood_info, METH_VARARGS},
{"get_peripheral_curve_datum", get_peripheral_curve_datum, METH_VARARGS},
    {NULL,                                      NULL                                                    }
};


void initSnapPeaC(void)
{
  PyObject *module;

    module = Py_InitModule("SnapPeaC", SnapPeaCMethods);
    //    SnapPea_Ack = PyErr_NewException("SnapPeaC.Acknowledge", NULL, NULL);
    //    Py_INCREF(SnapPea_Ack);
    //    PyModule_AddObject(module, "Acknowledge", SnapPea_Ack);
    SnapPea_Query = PyErr_NewException("SnapPeaC.Query", NULL, NULL);
    Py_INCREF(SnapPea_Query);
    PyModule_AddObject(module, "Query", SnapPea_Query);
    SnapPea_Fatal = PyErr_NewException("SnapPeaC.Fatal", NULL, NULL);
    Py_INCREF(SnapPea_Fatal);
    PyModule_AddObject(module, "Fatal", SnapPea_Fatal);
}

/* Helper functions */

static PyObject* int_array_to_PyList(int* array, int length)
{
  int i;
  PyObject* list;
  list = PyList_New(length);
  for (i = 0; i < length; i++)
    PyList_SetItem(list, i, PyInt_FromLong(array[i]));
  return list;
}

/* end helper functions */

static PyObject *get_gluing_data(PyObject *self, PyObject *args)
{
    long theTriangulation;
    int fill_first, fill_how_many = 0, num_cusps;
    Triangulation *filled_Triangulation = NULL;
    TriangulationData *data = NULL;
    int i,j,k; 
    PyObject *gluing_data, *neighbors, *gluings, *permutation, *tetrahedron_datum;

    if (!PyArg_ParseTuple(args, "li:get_gluing_data",
			  &theTriangulation, &fill_first))
    {
        PyErr_SetString(PyExc_TypeError, 
			"Invalid arguments passed to SnapPeaC.get_gluing_data.\n"
			"           Expected ((long)triangulation, (int)fill_first).");
        return NULL;
    }

    if (fill_first) {
      num_cusps = get_num_cusps((Triangulation *)theTriangulation);
      {
	Boolean fill_list[num_cusps];

	for (i = 0; i<num_cusps; i++)
	  if ( (fill_list[i] = cusp_is_fillable((Triangulation *)theTriangulation, i)) )
	    ++fill_how_many;

	if (fill_how_many > 0)
	  {
	    if (fill_how_many < num_cusps)
	      filled_Triangulation = fill_cusps((Triangulation *)theTriangulation,
						fill_list, "", FALSE);
	    else 
	      filled_Triangulation = fill_cusps((Triangulation *)theTriangulation,
						 NULL, "", TRUE);
	  }
      }
      triangulation_to_data(filled_Triangulation, &data);
    }
    else
      triangulation_to_data((Triangulation *)theTriangulation, &data);

    gluing_data = PyList_New(0);
       for (i = 0; i < data->num_tetrahedra; i++)
       {
           neighbors = PyTuple_New(4);
           gluings = PyTuple_New(4);
           for (j = 0; j < 4; j++)
           {
               permutation = PyTuple_New(4);
               for (k=0;k<4;k++)
                  PyTuple_SET_ITEM(permutation, k, PyInt_FromLong(data->tetrahedron_data[i].gluing[j][k]));
               PyTuple_SET_ITEM(neighbors, j, PyInt_FromLong(data->tetrahedron_data[i].neighbor_index[j]));
               PyTuple_SET_ITEM(gluings, j, permutation);
           }
           tetrahedron_datum = PyTuple_New(2);
           PyTuple_SET_ITEM(tetrahedron_datum, 0, neighbors);
           PyTuple_SET_ITEM(tetrahedron_datum, 1, gluings);
           PyList_Append(gluing_data, tetrahedron_datum);
		   Py_DECREF(tetrahedron_datum);
      }
    if (filled_Triangulation) free_triangulation(filled_Triangulation);
    if (data) free_triangulation_data(data);
    if ( PyErr_Occurred() ) return NULL;
    return gluing_data;
}

/*
   This function takes as argument a python list which is a
   Dowker-Thistlethwaite description of a knot, and returns a snapPea
   triangulation.  No consistency checking is done on the DT
   description.  That should be done in a python wrapper.
*/
extern Triangulation   *DT_int_to_triangulation(int aNumCrossings, int *aDTCode);

static PyObject *get_triangulation_from_DT(PyObject *self, PyObject *args)
{
    PyObject        *theDTlist, *aPyInt;
    Triangulation   *theTriangulation;
    int aNumCrossings, *aDTCode, i;

    if (!PyArg_ParseTuple(args, "O", &theDTlist)  || !PyList_Check(theDTlist) )
    {
        PyErr_SetString(PyExc_TypeError, "get_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    aNumCrossings = PyList_Size(theDTlist);
    aDTCode = malloc(aNumCrossings * sizeof(int));
    if (aDTCode == NULL) {
        PyErr_SetString(PyExc_RuntimeError, "Couldn't allocate aDTCode.");
	return NULL;
    }

    for (i=0; i < aNumCrossings; i++)
    {
        aPyInt = PyList_GetItem(theDTlist, i);
        if (aPyInt == NULL )
        {
            PyErr_SetString(PyExc_RuntimeError, "Invalid DT list.");
	    free(aDTCode);
            return NULL;
	}
        aDTCode[i] = (int)PyInt_AsLong(aPyInt);
    }
    
    theTriangulation = DT_int_to_triangulation(aNumCrossings, aDTCode);
    free(aDTCode);
    
    if (theTriangulation == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Could not create triangulation.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}

/*
   This function takes as arguments python objects comprising the equivalent
   of a SnapPea KLPProjection structure and returns a snapPea
   triangulation.  No consistency checking is done on the KLPProjection
   description.  That should be done in a python wrapper.
*/
extern Triangulation *triangulate_link_complement(KLPProjection *aLinkProjection);

static PyObject *get_triangulation_from_PythonKLP(PyObject *self, PyObject *args)
{
    KLPProjection P;
    Triangulation   *theTriangulation;
    PyObject *Pycrossings, *cr_dict;
    int i, neighbor, component;
    char *strand, *sign;

    if (!PyArg_ParseTuple(args, "iiiO", 
			  &P.num_crossings,
			  &P.num_free_loops,
			  &P.num_components,
			  &Pycrossings)  || !PyList_Check(Pycrossings) )
    {
        PyErr_SetString(PyExc_TypeError, "get_triangulation_from_PythonKLP() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    P.crossings = (KLPCrossing *)malloc(P.num_crossings * sizeof(KLPCrossing));
    if (P.crossings == NULL) {
        PyErr_SetString(PyExc_RuntimeError, "Couldn't allocate crossing table.");
	return NULL;
    }

    for (i=0; i < P.num_crossings; i++)
    {
        cr_dict = PyList_GetItem(Pycrossings, i);
        if ( !PyDict_Check(cr_dict) )
        {
            PyErr_SetString(PyExc_RuntimeError, "Invalid crossing list.");
	    free(P.crossings);
            return NULL;
	}
	neighbor = PyInt_AsLong(PyDict_GetItemString(cr_dict, "Xbackward_neighbor"));
	P.crossings[i].neighbor[KLPStrandX][KLPBackward] = P.crossings + neighbor; 

	neighbor =  PyInt_AsLong(PyDict_GetItemString(cr_dict, "Xforward_neighbor"));
	P.crossings[i].neighbor[KLPStrandX][KLPForward] = P.crossings + neighbor;

	neighbor = PyInt_AsLong(PyDict_GetItemString(cr_dict, "Ybackward_neighbor"));
	P.crossings[i].neighbor[KLPStrandY][KLPBackward] = P.crossings + neighbor;

	neighbor = PyInt_AsLong(PyDict_GetItemString(cr_dict, "Yforward_neighbor"));
	P.crossings[i].neighbor[KLPStrandY][KLPForward] = P.crossings + neighbor;

	strand = PyString_AS_STRING(PyDict_GetItemString(cr_dict, "Xbackward_strand"));
	P.crossings[i].strand[KLPStrandX][KLPBackward] = (strand[0]=='X' ? KLPStrandX : KLPStrandY);

	strand = PyString_AS_STRING(PyDict_GetItemString(cr_dict, "Xforward_strand"));
	P.crossings[i].strand[KLPStrandX][KLPForward] = (strand[0]=='X' ? KLPStrandX : KLPStrandY);

	strand = PyString_AS_STRING(PyDict_GetItemString(cr_dict, "Ybackward_strand"));
	P.crossings[i].strand[KLPStrandY][KLPBackward] = (strand[0]=='X' ? KLPStrandX : KLPStrandY);

	strand = PyString_AS_STRING(PyDict_GetItemString(cr_dict, "Yforward_strand"));
	P.crossings[i].strand[KLPStrandY][KLPForward] = (strand[0]=='X' ? KLPStrandX : KLPStrandY);

	sign = PyString_AS_STRING(PyDict_GetItemString(cr_dict, "sign"));
	P.crossings[i].handedness =  (sign[0]=='R' ? KLPHalfTwistCL : KLPHalfTwistCCL); 

	component = PyInt_AsLong(PyDict_GetItemString(cr_dict, "Xcomponent"));
	P.crossings[i].component[KLPStrandX] = component;

	component = PyInt_AsLong(PyDict_GetItemString(cr_dict, "Ycomponent"));
	P.crossings[i].component[KLPStrandY] = component;
    }

    theTriangulation = triangulate_link_complement(&P);
    free(P.crossings);

    if (theTriangulation == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Could not create triangulation.");
        return NULL;
    }

    /* The triangulation must have a name or SnapPea will segfault when
       trying to copy the triangulation. */
    set_triangulation_name(theTriangulation, "unnamed link");
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}

extern Triangulation *triangulate_punctured_torus_bundle(
                      LRFactorization *anLRFactorization);

/*
 *      If the manifold is hyperbolic (i.e. if the number of LR factors
 *      is at least two for an orientable bundle, or at least one for a
 *      nonorientable bundle), triangulates the complement and returns
 *      a pointer to it.  Otherwise returns NULL.
 */

Triangulation* easy_triangulate_punctured_torus_bundle(
               Boolean negative_determinant, 
	       Boolean negative_trace, char* LR_factors){
    Triangulation* manifold;
    LRFactorization*  glueing;

    glueing = alloc_LR_factorization(strlen(LR_factors));
    glueing->is_available = TRUE;
    glueing->negative_determinant = negative_determinant;
    glueing->negative_trace = negative_trace;
    strcpy(glueing->LR_factors, LR_factors);
    manifold =  triangulate_punctured_torus_bundle(glueing);
    free_LR_factorization(glueing);
    if ( PyErr_Occurred() ) return NULL;
    return manifold;
}

static PyObject *wrap_easy_triangulate_punctured_torus_bundle(PyObject *self, PyObject *args) {
    Triangulation *theTriangulation;
    Boolean  _arg0;
    Boolean  _arg1;
    char * _arg2;

    if(!PyArg_ParseTuple(args,"bbs:easy_triangulate_punctured_torus_bundle",&_arg0,&_arg1,&_arg2)) 
        return NULL;
    theTriangulation = (Triangulation *)easy_triangulate_punctured_torus_bundle(_arg0,_arg1,_arg2);
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}

// end t3m 

static PyObject *wrap_verify_my_malloc_usage(PyObject *self, PyObject *args)
{
    verify_my_malloc_usage();
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


static PyObject *wrap_get_triangulation(PyObject *self, PyObject *args)
{
    char            *theName;
    Triangulation   *theTriangulation;

    if (!PyArg_ParseTuple(args, "s", &theName))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    theTriangulation = get_triangulation(theName);
    
    if (theTriangulation == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Could not read named triangulation.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}


static PyObject *wrap_save_triangulation(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    char    *theName;

    if (!PyArg_ParseTuple(args, "ls", &theTriangulation, &theName))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_save_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    save_triangulation((Triangulation *)theTriangulation, theName);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


static PyObject *wrap_copy_triangulation(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    Triangulation   *theCopy;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_copy_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    copy_triangulation((Triangulation *)theTriangulation, &theCopy);
    
    if (theCopy == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Could not copy the triangulation.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theCopy);
}


static PyObject *wrap_get_cusped_census_manifold(PyObject *self, PyObject *args)
{
    int             theCensus;          //  5, 6, or 7
    int             theOrientability;   //  0 (false) or 1 (true)
    int             theIndex;           //  0 through #manifolds - 1
    char*        basePathName="";  // directory (including trailing os.sep) which contains CluspedCensusManifoldData
    Triangulation*  theTriangulation;

    if (!PyArg_ParseTuple(args, "siii", &basePathName, &theCensus, &theOrientability, &theIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_cusped_census_manifold() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    theTriangulation = GetCuspedCensusManifold(basePathName,
        theCensus,
        theOrientability ? oriented_manifold : nonorientable_manifold,
        theIndex);
    
    if (theTriangulation == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Could not load requested census manifold.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}


static PyObject *wrap_free_triangulation(PyObject *self, PyObject *args)
{
    long    theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_free_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    free_triangulation((Triangulation *) theTriangulation);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


static PyObject *wrap_get_triangulation_name(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    char    *theName;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_triangulation_name() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    //  We get a pointer to the actual name string (not a copy),
    //  so we won't need to free it.
    theName = get_triangulation_name((Triangulation *) theTriangulation);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("s", theName);
}


static PyObject *wrap_set_triangulation_name(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    char    *theName;

    if (!PyArg_ParseTuple(args, "ls", &theTriangulation, &theName))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_set_triangulation_name() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    set_triangulation_name((Triangulation *)theTriangulation, theName);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


static PyObject *wrap_get_triangulation_is_orientable(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    Orientability   theOrientability;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_triangulation_is_orientable() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    theOrientability = get_orientability((Triangulation *) theTriangulation);
    if (theOrientability != oriented_manifold
     && theOrientability != nonorientable_manifold)
    {
        PyErr_SetString(PyExc_RuntimeError, "Triangulation's orientability is unknown.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", theOrientability == oriented_manifold);
}


static PyObject *wrap_get_solution_type(PyObject *self, PyObject *args)
{
    long    theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_solution_type() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    switch (get_filled_solution_type((Triangulation *) theTriangulation))
    {
        case not_attempted:         return Py_BuildValue("s", "not attempted");
        case geometric_solution:    return Py_BuildValue("s", "all tetrahedra positively oriented");
        case nongeometric_solution: return Py_BuildValue("s", "contains negatively oriented tetrahedra");
        case flat_solution:         return Py_BuildValue("s", "contains flat tetrahedra");
        case degenerate_solution:   return Py_BuildValue("s", "contains degenerate tetrahedra");
        case other_solution:        return Py_BuildValue("s", "unrecognized solution type");
        case no_solution:           return Py_BuildValue("s", "no solution found");
        default:
            PyErr_SetString(PyExc_RuntimeError, "Bad solution type.");
            return NULL;
    }
}


static PyObject *wrap_get_num_tetrahedra(PyObject *self, PyObject *args)
{
    long    theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_num_tetrahedra() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", get_num_tetrahedra((Triangulation *) theTriangulation));
}


static PyObject *wrap_get_num_cusps(PyObject *self, PyObject *args)
{
    long    theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_num_cusps() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", get_num_cusps((Triangulation *) theTriangulation));
}


static PyObject *wrap_get_cusp_is_complete(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;
    Boolean theCuspIsComplete;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_cusp_is_complete() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    get_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
        NULL, &theCuspIsComplete, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", theCuspIsComplete);
}


static PyObject *wrap_get_cusp_is_orientable(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    int             theCuspIndex;
    CuspTopology    theCuspTopology;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_cusp_is_orientable() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    get_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
        &theCuspTopology, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

    //  Python scripts should never encounter Cusps of unknown CuspTopology.
    if (theCuspTopology != torus_cusp  &&  theCuspTopology != Klein_cusp)
    {
        PyErr_SetString(PyExc_RuntimeError, "Unknown cusp topology.");
        return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", theCuspTopology == torus_cusp);
}


static PyObject *wrap_get_cusp_m(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;
    double  m;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_cusp_m() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    get_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
                    NULL, NULL, &m, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("d", m);
}


static PyObject *wrap_get_cusp_l(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;
    double  l;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_cusp_l() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    get_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
                    NULL, NULL, NULL, &l, NULL, NULL, NULL, NULL, NULL, NULL);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("d", l);
}

static PyObject *wrap_get_holonomies(PyObject *self, PyObject *args)
{
    long       theTriangulation;
    int        theCuspIndex;
    int        meridian_precision;
    int        longitude_precision;
    int        theRecomputeFlag;
    Complex    snappea_meridian, snappea_longitude;
    PyObject   *py_meridian, *py_longitude;

    if (!PyArg_ParseTuple(args, "liiii", &theTriangulation, &theCuspIndex,
                &meridian_precision, &longitude_precision, &theRecomputeFlag))
    {
        PyErr_SetString(PyExc_TypeError,
        "wrap_get_holonomies() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if (theRecomputeFlag)
        (void) do_Dehn_filling((Triangulation *) theTriangulation);
    (void) get_holonomy((Triangulation *) theTriangulation, theCuspIndex,
			&snappea_meridian, &snappea_longitude,
			&meridian_precision, &longitude_precision);

    py_meridian = PyComplex_FromDoubles(snappea_meridian.real,
					snappea_meridian.imag);
    py_longitude = PyComplex_FromDoubles(snappea_longitude.real,
					 snappea_longitude.imag);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("OO", py_meridian, py_longitude);
}


static PyObject *wrap_get_angle_sums(PyObject *self, PyObject *args)
{
    long           theTriangulation;
    int            theRecomputeFlag;
    Triangulation *manifold;
    EdgeClass     *edge;
    PyObject      *theList,
                  *theAngle;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theRecomputeFlag))
    {
        PyErr_SetString(PyExc_TypeError,
        "wrap_get_angle_sums() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    manifold = (Triangulation *)theTriangulation;
    theList = PyList_New(0);
    if (theRecomputeFlag)
      compute_edge_angle_sums(manifold);

    for (   edge = manifold->edge_list_begin.next;
            edge != &manifold->edge_list_end;
            edge = edge->next)
    {
      theAngle = PyComplex_FromDoubles(edge->edge_angle_sum.real,
				       edge->edge_angle_sum.imag);
      PyList_Append(theList, theAngle);
      Py_DECREF(theAngle);
    }

    if ( PyErr_Occurred() ) return NULL;
    return theList; 
}

static PyObject *wrap_set_cusp_info(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;
    double  m,
            l;
    int     theRecomputeFlag;

    if (!PyArg_ParseTuple(args, "liddi", &theTriangulation, &theCuspIndex,
                &m, &l, &theRecomputeFlag))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_set_cusp_info() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    set_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
                    (m == 0.0 && l == 0.0), m, l);

    if (theRecomputeFlag)
        (void) do_Dehn_filling((Triangulation *) theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


static PyObject *wrap_cusp_is_fillable(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_cusp_is_fillable() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i",
        cusp_is_fillable((Triangulation *) theTriangulation, theCuspIndex));
}

static PyObject *wrap_remove_Dehn_fillings(PyObject *self, PyObject *args)
{
    long    theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_remove_Dehn_fillings() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    (void) remove_Dehn_fillings((Triangulation *) theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_volume(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    double  theVolume;
    int     precision;
    PyObject *ans;
    PyObject *py_volume;
    PyObject *py_precision;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_volume() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    theVolume = volume((Triangulation *) theTriangulation, &precision);

    ans = PyTuple_New(2);
    py_precision =  Py_BuildValue("i", precision);
    py_volume = Py_BuildValue("d", theVolume);
    PyTuple_SET_ITEM(ans, 0, py_volume);
    PyTuple_SET_ITEM(ans, 1, py_precision);
    if ( PyErr_Occurred() ) return NULL;
    return ans;
}


static PyObject *wrap_homology(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    AbelianGroup    *theHomology;
    PyObject        *theList;
    int             i;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_homology() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    theHomology = homology((Triangulation *) theTriangulation);
    
    if (theHomology == NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("");
    }

    compress_abelian_group(theHomology);

    theList = PyList_New(theHomology->num_torsion_coefficients);
    for (i = 0; i < theHomology->num_torsion_coefficients; i++)
        PyList_SetItem(theList, i, Py_BuildValue("l", theHomology->torsion_coefficients[i]));
    
    free_abelian_group(theHomology);

    if ( PyErr_Occurred() ) return NULL;
    return theList;
}


static PyObject *wrap_fill_cusp(PyObject *self, PyObject *args)
{
    long            theTri;
    Triangulation   *theTriangulation,
                    *theResult;
    int             theIndex,
                    n,
                    i;
    Boolean         *theFillArray;

    if (!PyArg_ParseTuple(args, "li", &theTri, &theIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_fill_cusp() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theTriangulation = (Triangulation *) theTri;
    
    if (theIndex < 0 || theIndex >= get_num_cusps(theTriangulation))
    {
        PyErr_SetString(PyExc_RuntimeError, "Cusp index out of bounds.");
        return NULL;
    }
    
    //  If the cusp isn't fillable, return the original Triangulation.
    if (cusp_is_fillable(theTriangulation, theIndex) == FALSE) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("l", theTri);
    }

    //  Otherwise do the filling, and return the result.
    n = get_num_cusps(theTriangulation);
    theFillArray = (Boolean *) malloc(n * sizeof(Boolean));
    if (theFillArray == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Couldn't allocate theFillArray.");
        return NULL;
    }
    for (i = 0; i < n; i++)
        theFillArray[i] = FALSE;
    theFillArray[theIndex] = TRUE;
    theResult = fill_cusps( theTriangulation,
                            theFillArray,
                            get_triangulation_name(theTriangulation),
                            FALSE);
    free(theFillArray);
    
    if (theResult == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Couldn't fill the requested cusps.");
        return NULL;
    }
    else {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("l", (long int) theResult);
    }
}



static PyObject *wrap_make_finite_triangulation(PyObject *self, PyObject *args)
{
    long            theTri;
    Triangulation   *theTriangulation,
                    *theResult;
    int             theIndex;

    if (!PyArg_ParseTuple(args, "l", &theTri, &theIndex))
    {
        PyErr_SetString(PyExc_TypeError, "make_finite_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theTriangulation = (Triangulation *) theTri;
    

    theResult = fill_cusps( theTriangulation,
                            NULL,
                            get_triangulation_name(theTriangulation),
                           TRUE);

    if (theResult == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Couldn't fill the all the cusps.");
        return NULL;
    }
    else {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("l", (long int) theResult);
    }
}


static PyObject *wrap_get_drillable_curves(PyObject *self, PyObject *args)
{
    long                    theTriangulation;
    int                     theMaxSegments,
                            theNumCurves,
                            i;
    DualOneSkeletonCurve    **theCurves;
    MatrixParity            theParity;
    Complex                 theCompleteLength,
                            theFilledLength;
    PyObject                *theReturnData,
                            *theCurveData;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theMaxSegments))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_drillable_curves() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    dual_curves((Triangulation *) theTriangulation,
                theMaxSegments,
                &theNumCurves,
                &theCurves);
    
    theReturnData = PyList_New(theNumCurves);
    
    for (i = 0; i < theNumCurves; i++)
    {
        get_dual_curve_info(theCurves[i],
                            &theCompleteLength,
                            &theFilledLength,
                            &theParity);
        
        theCurveData = Py_BuildValue("[i,d,d,d,d]",
                                    theParity == orientation_reversing,
                                    theFilledLength.real,
                                    theFilledLength.imag,
                                    theCompleteLength.real,
                                    theCompleteLength.imag);

        PyList_SetItem(theReturnData, i, theCurveData);
        //  PyList_SetItem() takes ownership of the reference to theCurveData,
        //  so we do NOT decrement theCurveData's reference count.
    }
    
    free_dual_curves(theNumCurves, theCurves);
    
    if ( PyErr_Occurred() ) return NULL;
    return theReturnData;
}


static PyObject *wrap_drill_curve(PyObject *self, PyObject *args)
{
    long                    theTriangulation;
    int                     theMaxSegments,
                            theIndex,
                            theNumCurves;
    Triangulation           *theResult;
    DualOneSkeletonCurve    **theCurves;

    if (!PyArg_ParseTuple(args, "lii", &theTriangulation, &theIndex, &theMaxSegments))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_drill_curve() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    dual_curves((Triangulation *) theTriangulation,
                theMaxSegments,
                &theNumCurves,
                &theCurves);
    
    if (theIndex < 0 || theIndex >= theNumCurves)
    {
        free_dual_curves(theNumCurves, theCurves);
        PyErr_SetString(PyExc_RuntimeError, "Curve index out of bounds.");
        return NULL;
    }
    
    theResult = drill_cusp( (Triangulation *) theTriangulation,
                            theCurves[theIndex],
                            get_triangulation_name((Triangulation *) theTriangulation) );

    free_dual_curves(theNumCurves, theCurves);
    
    if (theResult == NULL)
    {
        //  theResult will be NULL if the curve is boundary parallel.
        //  At present such curves aren't offered to the user,
        //  so NULL results shouldn't occur.  If we want to allow
        //  them in the future, replace "return NULL" with
        //
        //      return Py_BuildValue("l", (long int) 0);
        //
        PyErr_SetString(PyExc_RuntimeError, "Requested curve cannot be drilled because it's parallel to the boundary.");
        return NULL;
    }
    else {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("l", (long int) theResult);
    }
}


static PyObject *wrap_get_normal_surfaces(PyObject *self, PyObject *args)
{
    long                theTriangulation;
    NormalSurfaceList   *theSurfaceList;
    int                 theNumSurfaces,
                        i;
    PyObject            *theReturnData,
                        *theSurfaceDescription;
    char                theDescription[128];

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_normal_surfaces() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    switch (find_normal_surfaces((Triangulation *) theTriangulation, &theSurfaceList))
    {
        case func_OK:

            theNumSurfaces = number_of_normal_surfaces_on_list(theSurfaceList);
            
            theReturnData = PyList_New(theNumSurfaces);

            for (i = 0; i < theNumSurfaces; i++)
            {
                if (normal_surface_is_two_sided(theSurfaceList, i))
                    strcpy(theDescription, "2-sided ");
                else
                    strcpy(theDescription, "1-sided ");

                switch (normal_surface_Euler_characteristic(theSurfaceList, i))
                {
                    case 2:
                        assert(normal_surface_is_orientable(theSurfaceList, i) == TRUE);
                        strcat(theDescription, "sphere");
                        break;

                    case 1:
                        assert(normal_surface_is_orientable(theSurfaceList, i) == FALSE);
                        strcat(theDescription, "projective plane");
                        break;

                    case 0:
                        if (normal_surface_is_orientable(theSurfaceList, i))
                            strcat(theDescription, "torus");
                        else
                            strcat(theDescription, "Klein bottle");
                        break;
                    
                    default:    //  shouldn't ever happen
                        Py_DECREF(theReturnData);
                        free_normal_surfaces(theSurfaceList);
                        PyErr_SetString(PyExc_RuntimeError, "'Impossible' situation:  splitting surface has Euler characteristic not equal to 0, 1, or 2.");
                        return NULL;
                }

                theSurfaceDescription = Py_BuildValue("s", theDescription);

                PyList_SetItem(theReturnData, i, theSurfaceDescription);
                //  PyList_SetItem() takes ownership of the reference
                //  to theSurfaceDescription, so we do NOT decrement
                //  theSurfaceDescription's reference count.
            }

            free_normal_surfaces(theSurfaceList);

            if ( PyErr_Occurred() ) return NULL;
	    return theReturnData;
        
        case func_bad_input:
            if ( PyErr_Occurred() ) return NULL;
	    return Py_BuildValue("[s]", "splittings not yet available for closed manifolds");
        
        default:
            return Py_BuildValue("[s]", "???"); // should never occur
    }
}


static PyObject *wrap_split_along_normal_surface(PyObject *self, PyObject *args)
{
    long                theTriangulation;
    int                 theIndex;
    NormalSurfaceList   *theSurfaceList;
    int                 theNumSurfaces;
    Triangulation       *thePieces[2];
    PyObject            *thePieceList;
    char                *theOldName,
                        *theNewName;
    int                 i;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_split_along_normal_surface() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if (func_OK != find_normal_surfaces((Triangulation *) theTriangulation,
                                        &theSurfaceList))
    {
        PyErr_SetString(PyExc_RuntimeError, "Couldn't split along normal surface.  2-sided projective plane?");
        return NULL;
    }

    theNumSurfaces = number_of_normal_surfaces_on_list(theSurfaceList);
    assert(theIndex >= 0  &&  theIndex < theNumSurfaces);
    
    //  Even if split_along_normal_surface() returns func_bad_input,
    //  we can proceed normally.  thePieces[] will both be NULL.
    (void) split_along_normal_surface(theSurfaceList, theIndex, thePieces);
    
    free_normal_surfaces(theSurfaceList);
    
    thePieceList = PyList_New(0);

    theOldName = get_triangulation_name((Triangulation *) theTriangulation);
    for (i = 0; i < 2; i++)
        if (thePieces[i] != NULL)
        {
            theNewName = (char *) malloc((strlen(theOldName) + 3) * sizeof(char));
            strcpy(theNewName, theOldName);
            strcat(theNewName, i == 0 ? ".a" : ".b");
            set_triangulation_name(thePieces[i], theNewName);
            free(theNewName);

            PyList_Append(thePieceList, Py_BuildValue("l", (long) thePieces[i]));
        }

    if ( PyErr_Occurred() ) return NULL;
    return thePieceList;
}


static char *RelationToString(int *aSnapPeaRelation)
{
    int     *theLetter,
            theRelationLength,
            i;
    char    *theCString;
    
    //  How long is aSnapPeaRelation?
    theRelationLength = 0;
    for (   theLetter = aSnapPeaRelation;
            *theLetter != 0;
            theLetter++)
        theRelationLength++;
    
    //  Allocate memory for theCString.
    theCString = malloc(theRelationLength + 1);
    assert(theCString != NULL);

    //  Write the relation.
    for (   theLetter = aSnapPeaRelation, i = 0;
            *theLetter != 0;
            theLetter++, i++)
    {
        if (*theLetter > 0)
            theCString[i] = 'a' - 1 + *theLetter;
        else
            theCString[i] = 'A' - 1 - *theLetter;
    }
    assert(i == theRelationLength);
    theCString[i] = 0;
    
    if ( PyErr_Occurred() ) return NULL;
    return theCString;
}

static PyObject *wrap_fundamental_group(PyObject *self, PyObject *args)
{
    long                theTriangulation;
    int                 theSimplifyFlag,
                        theFillingsAffectGeneratorsFlag,
                        theMinimizeNumGeneratorsFlag;
    GroupPresentation   *theFundamentalGroup;

    if (!PyArg_ParseTuple(  args,
                            "liii",
                            &theTriangulation,
                            &theSimplifyFlag,
                            &theFillingsAffectGeneratorsFlag,
                            &theMinimizeNumGeneratorsFlag))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_fundamental_group() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    theFundamentalGroup = fundamental_group(
                        (Triangulation *) theTriangulation,
                        theSimplifyFlag,
                        theFillingsAffectGeneratorsFlag,
                        theMinimizeNumGeneratorsFlag);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long) theFundamentalGroup);
}

static PyObject *wrap_free_group_presentation(PyObject *self, PyObject *args)
{
    long                    theFundamentalGroup;

    if (!PyArg_ParseTuple(args, "l", &theFundamentalGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_free_group_presentation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    free_group_presentation((GroupPresentation *) theFundamentalGroup);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_fg_get_num_generators(PyObject *self, PyObject *args)
{
    long    theFundamentalGroup;

    if (!PyArg_ParseTuple(args, "l", &theFundamentalGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_fg_get_num_generators() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i",
        (fg_integer_fillings((GroupPresentation *) theFundamentalGroup) == TRUE) ?
        fg_get_num_generators((GroupPresentation *) theFundamentalGroup) :
        0);
}

static PyObject *wrap_fg_get_relations(PyObject *self, PyObject *args)
{
    long        theFundamentalGroup;
    int         theNumRelations,
                *theRawRelation,
                i;
    char        *theString;
    PyObject    *theRelationList;

    if (!PyArg_ParseTuple(args, "l", &theFundamentalGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_fg_get_relations() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    if (fg_integer_fillings((GroupPresentation *) theFundamentalGroup) == TRUE)
    {
        theNumRelations = fg_get_num_relations((GroupPresentation *) theFundamentalGroup);
        
        theRelationList = PyList_New(theNumRelations);

        for (i = 0; i < theNumRelations; i++)
        {
            theRawRelation = fg_get_relation((GroupPresentation *) theFundamentalGroup, i);
            theString = RelationToString(theRawRelation);
            fg_free_relation(theRawRelation);
            PyList_SetItem(theRelationList, i, Py_BuildValue("s", theString));
            free(theString);
        }
        
        if ( PyErr_Occurred() ) return NULL;
	return theRelationList;
    }
    else    
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("[]");
}

static PyObject *wrap_fg_representation(PyObject *self, PyObject *args)
{
    long                    theFundamentalGroup;
    int                     theWordLength;
    char                    *theWord;
    int                     *theSnapPeaWord;
    int                     i;
    FuncResult              theError;
    O31Matrix               theO31Matrix;
    MoebiusTransformation   theMoebiusTransformation;
    double                  (*o)[4];
    Complex                 (*s)[2];

    if (!PyArg_ParseTuple(  args,
                            "ls",
                            &theFundamentalGroup,
                            &theWord))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_fg_representation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    theWordLength   = strlen(theWord);
    theSnapPeaWord  = malloc((theWordLength + 1) * sizeof(int));
    
    for (i = 0; i < theWordLength; i++)
    {
        if (islower(theWord[i]))
            theSnapPeaWord[i] =  (theWord[i] - 'a' + 1);
        else if (isupper(theWord[i]))
            theSnapPeaWord[i] = -(theWord[i] - 'A' + 1);
    }
    theSnapPeaWord[theWordLength] = 0;

    theError = fg_word_to_matrix(
                        (GroupPresentation *) theFundamentalGroup,
                        theSnapPeaWord,
                        theO31Matrix,
                        &theMoebiusTransformation);
    
    free(theSnapPeaWord);
    
    if (theError == func_bad_input)
    {
        PyErr_SetString(PyExc_ValueError, "letter in word is not a valid generator");
        return NULL;
    }

    o = theO31Matrix;
    s = theMoebiusTransformation.matrix;
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue
    (   "(((dddd)(dddd)(dddd)(dddd)) (i(((dd)(dd))((dd)(dd)))))",

        o[0][0], o[0][1], o[0][2], o[0][3], 
        o[1][0], o[1][1], o[1][2], o[1][3], 
        o[2][0], o[2][1], o[2][2], o[2][3], 
        o[3][0], o[3][1], o[3][2], o[3][3], 
        
        theMoebiusTransformation.parity == orientation_preserving ? 0 : 1,
        s[0][0].real, s[0][0].imag,    s[0][1].real, s[0][1].imag,
        s[1][0].real, s[1][0].imag,    s[1][1].real, s[1][1].imag
    );
}

static PyObject *wrap_fg_peripheral_curves(PyObject *self, PyObject *args)
{
    long                    theFundamentalGroup;
    int                     theNumCusps,
                            i;
    int                     *theMeridian,
                            *theLongitude;
    char                    *theMString,
                            *theLString;
    PyObject                *thePeripheralCurves;

    if (!PyArg_ParseTuple(args, "l", &theFundamentalGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_fg_peripheral_curves() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    theNumCusps = fg_get_num_cusps((GroupPresentation *) theFundamentalGroup);

    thePeripheralCurves = PyList_New(theNumCusps);
    
    for (i = 0; i < theNumCusps; i++)
    {
        theMeridian  = fg_get_meridian ((GroupPresentation *) theFundamentalGroup, i);
        theLongitude = fg_get_longitude((GroupPresentation *) theFundamentalGroup, i);
        
        theMString = RelationToString(theMeridian );
        theLString = RelationToString(theLongitude);
        
        fg_free_relation(theMeridian);
        fg_free_relation(theLongitude);
        
        PyList_SetItem(thePeripheralCurves, i,
            Py_BuildValue("[ss]", theMString, theLString));
        
        free(theMString);
        free(theLString);
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return thePeripheralCurves;
}


static PyObject *wrap_core_geodesic(PyObject *self, PyObject *args)
{
    long        theTriangulation;
    int         theCuspIndex;
    int         theSingularityIndex;
    Complex     theComplexLength,
                theHolonomy,
                theTraceSquared,
                theTrace,
                theEigenvalue;
    int         thePrecision;
    PyObject    *theDictionary,
                *theObject;

    static Complex  zero = { 0.0, 0.0},
                    one  = { 1.0, 0.0},
                    two  = { 2.0, 0.0};

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_core_geodesic() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    core_geodesic(  (Triangulation *) theTriangulation,
                    theCuspIndex,
                    &theSingularityIndex,
                    &theComplexLength,
                    &thePrecision);
    
    if (theSingularityIndex > 0)
    {
        //  complex_length.c explains the relationship between
        //  holonomy, complex length and trace.  It uses the
        //  variable 'k' for theHolonomy.
        theHolonomy     = complex_exp(theComplexLength);
        theTraceSquared = complex_plus (
                            complex_plus(
                                theHolonomy,
                                complex_div(one, theHolonomy)),
                            two);
        theTrace        = complex_sqrt(theTraceSquared);
        theEigenvalue   = complex_exp(complex_real_mult(0.5, theComplexLength));
    }
    else
    {
        theComplexLength    = zero;
        theHolonomy         = zero;
        theTraceSquared     = zero;
        theTrace            = zero;
        theEigenvalue       = zero;
    }

    theDictionary = PyDict_New();

    theObject = Py_BuildValue("i", theSingularityIndex);
    PyDict_SetItemString(theDictionary, "singularity index", theObject);
    Py_DECREF(theObject);

    theObject = Py_BuildValue("i", thePrecision);
    PyDict_SetItemString(theDictionary, "precision", theObject);
    Py_DECREF(theObject);

    theObject = PyComplex_FromDoubles(theComplexLength.real, theComplexLength.imag  );
    PyDict_SetItemString(theDictionary, "complex length", theObject);
    Py_DECREF(theObject);

    theObject = PyComplex_FromDoubles(theHolonomy.real,      theHolonomy.imag    );
    PyDict_SetItemString(theDictionary, "holonomy", theObject);
    Py_DECREF(theObject);

    theObject = PyComplex_FromDoubles(theTraceSquared.real,  theTraceSquared.imag);
    PyDict_SetItemString(theDictionary, "trace squared", theObject);
    Py_DECREF(theObject);

    theObject = PyComplex_FromDoubles(theTrace.real,         theTrace.imag       );
    PyDict_SetItemString(theDictionary, "trace", theObject);
    Py_DECREF(theObject);

    theObject = PyComplex_FromDoubles(theEigenvalue.real,    theEigenvalue.imag  );
    PyDict_SetItemString(theDictionary, "eigenvalue", theObject);
    Py_DECREF(theObject);
    
    if ( PyErr_Occurred() ) return NULL;
    return theDictionary;
}

static PyObject *wrap_change_peripheral_curves(PyObject *self, PyObject *args)
{
  long     theTriangulation;
  int      theCuspIndex, a,b,c,d, numcusps = get_num_cusps((Triangulation *)theTriangulation);
  MatrixInt22 *theMatrices;
  FuncResult theFuncResult;

  if (!PyArg_ParseTuple(args, "liiiii:change_peripheral_curves",
			&theTriangulation, &theCuspIndex, &a, &b, &c, &d))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_change_peripheral_curves() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

  theMatrices = (MatrixInt22 *)malloc(numcusps*sizeof(MatrixInt22));
  {
    int i,j,n;
    for (n=0; n<numcusps; n++)
        for (i = 0; i < 2; i++)
            for (j = 0; j < 2; j++)
                theMatrices[n][i][j] = (i == j);
  }

  theMatrices[theCuspIndex][0][0] = a;
  theMatrices[theCuspIndex][0][1] = b;
  theMatrices[theCuspIndex][1][0] = c;
  theMatrices[theCuspIndex][1][1] = d;

  theFuncResult = change_peripheral_curves((Triangulation *) theTriangulation, theMatrices);

  free(theMatrices);
  if (theFuncResult == func_bad_input){
    PyErr_SetString(PyExc_ValueError, 
       "The change-of-basis matrix must have determinant 1.");
      return NULL;
    }

  if ( PyErr_Occurred() ) return NULL;
  return Py_BuildValue("");
}

static PyObject *wrap_shortest_curves_become_meridians(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_shortest_curves_become_meridians() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    install_shortest_bases((Triangulation *)theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_current_fillings_become_meridians(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_current_fillings_become_meridians() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    install_current_curve_bases((Triangulation *)theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


static PyObject *wrap_basic_simplification(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_basic_simplification() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    basic_simplification((Triangulation *)theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_randomize_triangulation(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_randomize_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    randomize_triangulation((Triangulation *)theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_reorient(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_reorient() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    reorient((Triangulation *)theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_proto_canonize(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_proto_canonize() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    switch (proto_canonize((Triangulation *)theTriangulation))
    {
        case func_OK:
            if (is_canonical_triangulation((Triangulation *)theTriangulation) == FALSE)
                uAcknowledge("The canonical cell decomposition contains cells other than tetrahedra.  Such cells have been arbitrarily subdivided into tetrahedra.");
            break;
        
        case func_failed:
            uAcknowledge("Only hyperbolic manifolds have canonical decompositions.");
            break;
        
        default:
            //  Should not occur.
            PyErr_SetString(PyExc_RuntimeError, "Unknown error in proto_canonize().");
            return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_is_canonical_triangulation(PyObject *self, PyObject *args)
{
    long        theTriangulation;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_is_canonical_triangulation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", is_canonical_triangulation((Triangulation *)theTriangulation));
}

static PyObject *wrap_are_isometric(PyObject *self, PyObject *args)
{
    long            FirstTriangulation, SecondTriangulation;
    Boolean         are_isometric;
    FuncResult theFuncResult;

    if (!PyArg_ParseTuple(args, "ll", &FirstTriangulation, &SecondTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_are_isometric() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    theFuncResult = compute_isometries((Triangulation *)FirstTriangulation, 
				       (Triangulation *)SecondTriangulation, 
				       &are_isometric, NULL, NULL);
    if (theFuncResult == func_bad_input){
      PyErr_SetString(PyExc_ValueError, 
            "wrap_are_isometric() in SnapPeaC.c received Dehn filling coeffs not rel. prime integers.");
      return NULL;
    }

    if (theFuncResult == func_failed){
      PyErr_SetString(PyExc_RuntimeError, 
            "wrap_are_isometric() in SnapPeaC.c could not decide if the manifolds are isometric.");
      return NULL;
    }

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", (int) are_isometric);
}


static PyObject *wrap_symmetry_group(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    SymmetryGroup   *symmetry_group_of_manifold,
                    *symmetry_group_of_link;
    Triangulation   *symmetric_triangulation;
    Boolean         is_full_group;

    if (!PyArg_ParseTuple(args, "l", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    symmetry_group_of_manifold  = NULL;
    symmetry_group_of_link      = NULL;
    symmetric_triangulation     = NULL;
    is_full_group               = FALSE;

    //  In the rare event that compute_symmetry_group() fails,
    //  the various pointers will be left as NULL.
    
    (void) compute_symmetry_group(  (Triangulation *)theTriangulation,
                                    &symmetry_group_of_manifold,
                                    &symmetry_group_of_link,
                                    &symmetric_triangulation,
                                    &is_full_group);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("[llli]",
                            (long) symmetry_group_of_manifold,
                            (long) symmetry_group_of_link,
                            (long) symmetric_triangulation,
                            is_full_group);
}

static PyObject *wrap_free_symmetry_group(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_free_symmetry_group() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    free_symmetry_group((SymmetryGroup *) theSymmetryGroup);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_symmetry_group_order(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_order() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", symmetry_group_order((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_is_abelian(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_is_abelian() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", (int) symmetry_group_is_abelian((SymmetryGroup *) theSymmetryGroup, NULL));
}

static PyObject *wrap_symmetry_group_abelian_description(PyObject *self, PyObject *args)
{
    long            theSymmetryGroup;
    AbelianGroup    *theAbelianDescription;
    PyObject        *theList;
    int             i;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_abelian_description() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if (symmetry_group_is_abelian(  (SymmetryGroup *) theSymmetryGroup,
                                    &theAbelianDescription)
        != TRUE)
    {
        PyErr_SetString(PyExc_TypeError, "Abelian desciptions aren't available for nonabelian groups.  Use SymmetryGroup.is_abelian() to test.");
        return NULL;
    }

    theList = PyList_New(theAbelianDescription->num_torsion_coefficients);
    for (i = 0; i < theAbelianDescription->num_torsion_coefficients; i++)
        PyList_SetItem(theList, i, Py_BuildValue("l", theAbelianDescription->torsion_coefficients[i]));
    
    if ( PyErr_Occurred() ) return NULL;
    return theList;
}

static PyObject *wrap_symmetry_group_is_dihedral(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_is_dihedral() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", (int) symmetry_group_is_dihedral((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_is_polyhedral(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_is_polyhedral() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", (int) symmetry_group_is_polyhedral(
        (SymmetryGroup *) theSymmetryGroup, NULL, NULL, NULL, NULL));
}

static PyObject *wrap_symmetry_group_polyhedral_description(PyObject *self, PyObject *args)
{
    long        theSymmetryGroup;
    Boolean     is_binary_group;
    int         p,
                q,
                r;
    PyObject    *theDictionary,
                *theObject;
    char        theName[64];

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_polyhedral_description() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    if (symmetry_group_is_polyhedral(   (SymmetryGroup *) theSymmetryGroup,
                                        &is_binary_group,
                                        &p,
                                        &q,
                                        &r
        ) != TRUE)
    {
        PyErr_SetString(PyExc_ValueError, "Symmetry group is not a polyhedral group, and therefore does not have a polyhedral description.  Use SymmetryGroup.is_polyhedral() to test.");
        return NULL;
    }

    //  Use the fact that the SnapPea kernel generates the (p,q,r)
    //  in ascending order.
    
    assert(p == 2);
    
    switch(q)
    {
        case 2:
            //  Plain dihedral groups are best handled
            //  in the specialized dihedral group code,
            //  which puts the elements into a natural order.
            assert(is_binary_group == TRUE);
            
            sprintf(theName, "binary dihedral group <2,2,%d>", r);

            break;
    
        case 3:

            strcpy(theName, is_binary_group ? "binary " : "");
    
            switch (r)
            {
                case 3:  strcat(theName, "tetrahedral group");  break;
                case 4:  strcat(theName,  "octahedral group");  break;
                case 5:  strcat(theName, "icosahedral group");  break;
                default:  assert(FALSE);
            }
            break;
        
        default:  assert(FALSE);
    }

    theDictionary = PyDict_New();
    
    theObject = Py_BuildValue("i", is_binary_group ? 1 : 0);
    PyDict_SetItemString(   theDictionary,  "is binary",    theObject);
    Py_DECREF(theObject);

    theObject = Py_BuildValue("(iii)", p, q, r);
    PyDict_SetItemString(   theDictionary,  "pqr",          theObject);
    Py_DECREF(theObject);

    theObject = Py_BuildValue("s", theName);
    PyDict_SetItemString(   theDictionary,  "name",         theObject);
    Py_DECREF(theObject);
    
    if ( PyErr_Occurred() ) return NULL;
    return theDictionary;
}

static PyObject *wrap_symmetry_group_is_S5(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_is_S5() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", (int) symmetry_group_is_S5((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_is_direct_product(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_is_direct_product() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", (int) symmetry_group_is_direct_product((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_factor(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;
    int     theFactorIndex;

    if (!PyArg_ParseTuple(args, "li", &theSymmetryGroup, &theFactorIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_factor() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    if (theFactorIndex != 0 && theFactorIndex != 1)
    {
        PyErr_SetString(PyExc_ValueError, "wrap_symmetry_group_factor() provides only two factors, indexed 0 and 1, which may be factored recursively if necessary.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) get_symmetry_group_factor((SymmetryGroup *) theSymmetryGroup, theFactorIndex));
}

static PyObject *wrap_symmetry_group_is_amphicheiral(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_is_amphicheiral() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", symmetry_group_is_amphicheiral((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_invertible_knot(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_invertible_knot() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("i", symmetry_group_invertible_knot((SymmetryGroup *) theSymmetryGroup));
}


static PyObject *wrap_symmetry_group_commutator_subgroup(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_commutator_subgroup() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) get_commutator_subgroup((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_abelianization(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_abelianization() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) get_abelianization((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_center(PyObject *self, PyObject *args)
{
    long    theSymmetryGroup;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_center() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) get_center((SymmetryGroup *) theSymmetryGroup));
}

static PyObject *wrap_symmetry_group_presentation(PyObject *self, PyObject *args)
{
    long                        theSymmetryGroup;
    SymmetryGroupPresentation   *thePresentation;
    PyObject                    *theRelations,
                                *theRelation,
                                *theDictionary,
                                *theObject;
    int                         theNumGenerators,
                                theNumRelations,
                                theNumFactors,
                                theGenerator,
                                thePower,
                                i,
                                j;

    if (!PyArg_ParseTuple(args, "l", &theSymmetryGroup))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_symmetry_group_presentation() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    thePresentation = get_symmetry_group_presentation((SymmetryGroup *) theSymmetryGroup);

    theNumGenerators = sg_get_num_generators(thePresentation);
    theNumRelations  = sg_get_num_relations(thePresentation);
    
    theRelations = PyList_New(theNumRelations);
    for (i = 0; i < theNumRelations; i++)
    {
        theNumFactors = sg_get_num_factors(thePresentation, i);
        theRelation = PyList_New(theNumFactors);
        for (j = 0; j < theNumFactors; j++)
        {
            sg_get_factor(thePresentation, i, j, &theGenerator, &thePower);
            PyList_SetItem(theRelation, j, Py_BuildValue("(ii)", theGenerator, thePower));
        }
        PyList_SetItem(theRelations, i, theRelation);
    }

    free_symmetry_group_presentation(thePresentation);

    theDictionary = PyDict_New();

    theObject = Py_BuildValue("i", theNumGenerators);
    PyDict_SetItemString(theDictionary, "number of generators", theObject);
    Py_DECREF(theObject);

    theObject = Py_BuildValue("i", theNumRelations);
    PyDict_SetItemString(theDictionary, "number of relations",  theObject);
    Py_DECREF(theObject);

    PyDict_SetItemString(theDictionary, "relations",            theRelations);
    Py_DECREF(theRelations);
    
    if ( PyErr_Occurred() ) return NULL;
    return theDictionary;
}


static PyObject *wrap_tet_shapes(PyObject *self, PyObject *args)
{
    long        theTriangulation;
    int         theFixedAlignment;
    int         theNumTetrahedra,
                i;
    PyObject    *theList,
                *theDictionary,
                *theObject;
    double      theShapeRectReal,
                theShapeRectImag,
                theShapeLogReal,
                theShapeLogImag;
    int         thePrecisionRectReal,
                thePrecisionRectImag,
                thePrecisionLogReal,
                thePrecisionLogImag;
    Boolean     theGeometricFlag;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theFixedAlignment))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_tet_shapes() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    theNumTetrahedra = get_num_tetrahedra((Triangulation *)theTriangulation);

    theList = PyList_New(theNumTetrahedra);

    for (i = 0; i < theNumTetrahedra; i++)
    {
        get_tet_shape(  (Triangulation *)theTriangulation,
                        i,
                        theFixedAlignment,
                        &theShapeRectReal,
                        &theShapeRectImag,
                        &theShapeLogReal,
                        &theShapeLogImag,
                        &thePrecisionRectReal,
                        &thePrecisionRectImag,
                        &thePrecisionLogReal,
                        &thePrecisionLogImag,
                        &theGeometricFlag);
        
        theDictionary = PyDict_New();

        theObject = PyComplex_FromDoubles(theShapeRectReal, theShapeRectImag);
        PyDict_SetItemString(theDictionary, "shape rect", theObject);
        Py_DECREF(theObject);

        theObject = PyComplex_FromDoubles(theShapeLogReal,  theShapeLogImag );
        PyDict_SetItemString(theDictionary, "shape log",  theObject);
        Py_DECREF(theObject);

        theObject = Py_BuildValue("i", thePrecisionRectReal);
        PyDict_SetItemString(theDictionary, "precision rect real", theObject);
        Py_DECREF(theObject);

        theObject = Py_BuildValue("i", thePrecisionRectImag);
        PyDict_SetItemString(theDictionary, "precision rect imag", theObject);
        Py_DECREF(theObject);

        theObject = Py_BuildValue("i", thePrecisionLogReal);
        PyDict_SetItemString(theDictionary, "precision log real",  theObject);
        Py_DECREF(theObject);

        theObject = Py_BuildValue("i", thePrecisionLogImag);
        PyDict_SetItemString(theDictionary, "precision log imag",  theObject);
        Py_DECREF(theObject);

        theObject = Py_BuildValue("i", theGeometricFlag);
        PyDict_SetItemString(theDictionary, "is geometric",  theObject);
        Py_DECREF(theObject);
        
        PyList_SetItem(theList, i, theDictionary);
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return theList;
}

static PyObject *wrap_set_tet_shapes(PyObject *self, PyObject *args)
{
  long            theTriangulation;
  int             theNumTetrahedra;
  PyObject       *theShapeList;

    if (!PyArg_ParseTuple(args, "lO", &theTriangulation, &theShapeList))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_tet_shapes() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    theNumTetrahedra = get_num_tetrahedra((Triangulation *)theTriangulation);
    if (theNumTetrahedra != PySequence_Length(theShapeList))
        PyErr_SetString(PyExc_ValueError, "wrap_tet_shapes() in SnapPeaC.c received the wrong number of tetrahedron shapes.");

    {
        Complex  theShapes[theNumTetrahedra];
	PyObject *aShape;
	int i;
	for (i=0; i<theNumTetrahedra; i++)
        {
	    aShape = PySequence_GetItem(theShapeList, i);
	    if ( PyComplex_Check(aShape) )
	    {
	        theShapes[i].real = PyComplex_RealAsDouble(aShape);
		theShapes[i].imag = PyComplex_ImagAsDouble(aShape);
		Py_DECREF(aShape);
	    }
	    else
	    {
	        PyErr_SetString(PyExc_TypeError, "shapes must be complex numbers.");
		Py_DECREF(aShape);
		return NULL;
	    }
	}
	set_tet_shapes((Triangulation *)theTriangulation, theShapes);
	if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("");
    }
}

static PyObject *wrap_set_target_holonomy(PyObject *self, PyObject *args)
{
    long     theTriangulation;
    int      theCuspIndex;
    PyObject *theTarget;
    int      theRecomputeFlag;
    Cusp     *cusp;

    if (!PyArg_ParseTuple(args, "liOi", &theTriangulation, &theCuspIndex,
                &theTarget, &theRecomputeFlag))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_set_target_holonomy() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    cusp = find_cusp( (Triangulation *) theTriangulation, theCuspIndex);
    cusp->target_holonomy.real = PyComplex_RealAsDouble(theTarget);
    cusp->target_holonomy.imag = PyComplex_ImagAsDouble(theTarget);

    if (theRecomputeFlag)
        (void) do_Dehn_filling((Triangulation *) theTriangulation);

    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_Dirichlet(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    WEPolyhedron    *theDirichletDomain;
    int             theCentroidAtOriginFlag,
                    theMaximizeInjRadiusFlag;
    double          displacement[3];

    if (!PyArg_ParseTuple(args, "liiddd",
            &theTriangulation,
            &theCentroidAtOriginFlag,
            &theMaximizeInjRadiusFlag,
            &displacement[0],
            &displacement[1],
            &displacement[2]))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }

    theDirichletDomain = Dirichlet_with_displacement(
        (Triangulation *) theTriangulation, //  manifold
        displacement,                       //  displacement
        1e-8,                               //  vertex epsilon
        theCentroidAtOriginFlag,            //  centroid_at_origin
        Dirichlet_keep_going,               //  DirichletInteractivity
        theMaximizeInjRadiusFlag);          //  maximize_injectivity_radius
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long) theDirichletDomain);
}

static PyObject *wrap_free_Dirichlet_domain(PyObject *self, PyObject *args)
{
    long    theDirichletDomain;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_free_Dirichlet_domain() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    if (theDirichletDomain != 0)
        free_Dirichlet_domain((WEPolyhedron *) theDirichletDomain);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_Dirichlet_num_vertices(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_num_vertices() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;
    
    if (theDD != NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("i", theDD->num_vertices);
    }
    else {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("i", 0);
    }
}

static PyObject *wrap_Dirichlet_num_edges(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_num_edges() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;
    
    if (theDD != NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("i", theDD->num_edges);
    }
    else {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("i", 0);
    }
}

static PyObject *wrap_Dirichlet_num_faces(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_num_faces() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;
    
    if (theDD != NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("i", theDD->num_faces);
    }
    else {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("i", 0);
    }
}

static PyObject *wrap_Dirichlet_vertices(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;
    PyObject        *theList;
    WEVertex        *theVertex;
    int             theCount;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_vertices() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;

    if (theDD == NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("[]");
    }

    theList = PyList_New(theDD->num_vertices);

    for (theVertex = theDD->vertex_list_begin.next, theCount = 0;
         theVertex != &theDD->vertex_list_end;
         theVertex = theVertex->next, theCount++)
    {
        PyList_SetItem(theList, theCount, 
            Py_BuildValue("(fff)", theVertex->x[1], theVertex->x[2], theVertex->x[3]));
    }
    assert(theCount == theDD->num_vertices);
    
    if ( PyErr_Occurred() ) return NULL;
    return theList;
}

static PyObject *wrap_Dirichlet_faces(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;
    WEVertex        *theVertex,
                    *theCandidateVertex;
    WEEdge          *theEdge;
    WEFace          *theFace;
    PyObject        *theList,
                    *theVertexList;
    int             theFaceCount,
                    theVertexCount,
                    theVertexIndex;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_faces() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;

    if (theDD == NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("[]");
    }

    theList = PyList_New(theDD->num_faces);

    for (theFace = theDD->face_list_begin.next, theFaceCount = 0;
         theFace != &theDD->face_list_end;
         theFace = theFace->next, theFaceCount++)
    {
        theVertexList   = PyList_New(theFace->num_sides);
        theVertexCount  = 0;

        //  Enumerate the indices of the vertices, travelling counterclockwise.
        theEdge = theFace->some_edge;
        do
        {
            //  Find the more clockwise of theEdge's two endpoints.
            if (theEdge->f[left] == theFace)
                theVertex = theEdge->v[tip];
            else
                theVertex = theEdge->v[tail];

            //  The vertices don't have explicit indices, so we have
            //  to figure out on the fly where each vertex is on the
            //  list.  This isn't very efficient, but this isn't
            //  a time-critical routine.
            theVertexIndex = 0;
            theCandidateVertex = theDD->vertex_list_begin.next;
            while (theCandidateVertex != theVertex)
            {
                theVertexIndex++;
                theCandidateVertex = theCandidateVertex->next;
            }
            PyList_SetItem( theVertexList,
                            theVertexCount++,
                            Py_BuildValue("i", theVertexIndex));
    
            //  Move on to the next edge.
            if (theEdge->f[left] == theFace)
                theEdge = theEdge->e[tip][left];
            else
                theEdge = theEdge->e[tail][right];
    
        } while (theEdge != theFace->some_edge);
        assert(theVertexCount == theFace->num_sides);
        
        PyList_SetItem(theList, theFaceCount, theVertexList);
    }
    assert(theFaceCount == theDD->num_faces);

    if ( PyErr_Occurred() ) return NULL;
    return theList;
}

static PyObject *wrap_Dirichlet_face_colors(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;
    PyObject        *theList;
    WEFace          *theFace;
    int             theCount;
    double          h,
                    r,
                    g,
                    b;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_face_colors() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;

    if (theDD == NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("[]");
    }

    theList = PyList_New(theDD->num_faces);

    for (theFace = theDD->face_list_begin.next, theCount = 0;
         theFace != &theDD->face_list_end;
         theFace = theFace->next, theCount++)
    {
        //  Determine the face color.
        
        //  Uh-oh.  I know I want
        //
        //      hue         = theFace->f_class->hue;
        //      saturation  = 1.0;
        //      value       = 0.75;
        //
        //  but I have no idea how to convert HSV to RGB on unix.
        //  For now, let's just fudge it.
        
        h = 3.0 * theFace->f_class->hue;
        if (h < 1.0)
        {
            r = 0.25 + 0.75*(1.0 - h);
            g = 0.25 + 0.75*(h - 0.0);
            b = 0.25;
        }
        else if (h < 2.0)
        {
            r = 0.25;
            g = 0.25 + 0.75*(2.0 - h);
            b = 0.25 + 0.75*(h - 1.0);
        }
        else
        {
            r = 0.25 + 0.75*(h - 2.0);
            g = 0.25;
            b = 0.25 + 0.75*(3.0 - h);
        }

        PyList_SetItem(theList, theCount,
            Py_BuildValue("(fff)", r, g, b));
    }
    assert(theCount == theDD->num_faces);
    
    if ( PyErr_Occurred() ) return NULL;
    return theList;
}

static PyObject *wrap_Dirichlet_face_pairings(PyObject *self, PyObject *args)
{
    long            theDirichletDomain;
    WEPolyhedron    *theDD;
    PyObject        *theList;
    WEFace          *theFace;
    int             theCount;

    if (!PyArg_ParseTuple(args, "l", &theDirichletDomain))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_Dirichlet_face_pairings() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    theDD = (WEPolyhedron *) theDirichletDomain;

    if (theDD == NULL) {
        if ( PyErr_Occurred() ) return NULL;
	return Py_BuildValue("[]");
    }

    theList = PyList_New(theDD->num_faces);

    for (theFace = theDD->face_list_begin.next, theCount = 0;
         theFace != &theDD->face_list_end;
         theFace = theFace->next, theCount++)
    {
        PyList_SetItem(theList, theCount,
            Py_BuildValue("(ffff)(ffff)(ffff)(ffff)",
            (*(theFace->group_element))[0][0],
            (*(theFace->group_element))[0][1],
            (*(theFace->group_element))[0][2],
            (*(theFace->group_element))[0][3],
            (*(theFace->group_element))[1][0],
            (*(theFace->group_element))[1][1],
            (*(theFace->group_element))[1][2],
            (*(theFace->group_element))[1][3],
            (*(theFace->group_element))[2][0],
            (*(theFace->group_element))[2][1],
            (*(theFace->group_element))[2][2],
            (*(theFace->group_element))[2][3],
            (*(theFace->group_element))[3][0],
            (*(theFace->group_element))[3][1],
            (*(theFace->group_element))[3][2],
            (*(theFace->group_element))[3][3]
            ));
    }
    assert(theCount == theDD->num_faces);
    
    if ( PyErr_Occurred() ) return NULL;
    return theList;
}


static PyObject *wrap_get_current_cusp_shape(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;
    int precision;
    Complex cusp_shape;
     PyObject *ans;
     PyObject *py_shape;
     PyObject *py_precision;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_current_cusp_shape() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    get_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
        NULL, NULL, NULL, NULL, NULL, &cusp_shape, NULL, &precision, NULL, NULL);

    ans = PyTuple_New(2);
    py_precision =  Py_BuildValue("i", precision);
    py_shape = PyComplex_FromDoubles(  cusp_shape.real, cusp_shape.imag );
    PyTuple_SET_ITEM(ans, 0, py_shape);
    PyTuple_SET_ITEM(ans, 1, py_precision);
    if ( PyErr_Occurred() ) return NULL;
    return ans;
}

static PyObject *wrap_get_current_cusp_modulus(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int     theCuspIndex;
    int precision;
    Complex cusp_shape;
     PyObject *ans;
     PyObject *py_shape;
     PyObject *py_precision;

    if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theCuspIndex))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_get_current_cusp_shape() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    
    get_cusp_info(  (Triangulation *) theTriangulation, theCuspIndex,
        NULL, NULL, NULL, NULL, NULL, NULL, NULL, &precision, NULL, &cusp_shape);

    ans = PyTuple_New(2);
    py_precision =  Py_BuildValue("i", precision);
    py_shape = PyComplex_FromDoubles(  cusp_shape.real, cusp_shape.imag );
    PyTuple_SET_ITEM(ans, 0, py_shape);
    PyTuple_SET_ITEM(ans, 1, py_precision);
    if ( PyErr_Occurred() ) return NULL;
    return ans;
}



/* Format of get_gluing_equations:
 *
 *        a b c  d e f  ... = 2 pi i
 *
 *  means
 *       
 *     a*log(z0) + b*log(1/(1-z0)) + c*log((z0-1)/z) + d*log(z1) +... = 2 pi i
 *
 *  for edge equations, and (same) = 1 for cusp equations
 * 
 *  In terms of the tetrahedra, a is the invariant of the edge (2,3), b the invariant
 * of the edge (0,2) and c is the invariant of the edge (1,2).  See kernel_code/edge_classes.c
 * for a detailed account of the convention.
 */ 


static PyObject *get_gluing_equations(PyObject *self, PyObject *args)
{
    long theTriangulation;
    Triangulation*  manifold = NULL;
    PyObject *gluing_equations, *equation;


    int             *eqn,  i, T;
    EdgeClass       *edge;
    PositionedTet   ptet0, ptet;

    if (!PyArg_ParseTuple(args, "l:get_gluing_equations",
			  &theTriangulation))      {
        PyErr_SetString(PyExc_TypeError, 
			"Invalid arguments passed to SnapPeaC.get_gluing_equations.\n"
			"           Expected ((long)triangulation).");
        return NULL;
    }


    manifold = (Triangulation*) theTriangulation;
    T = manifold -> num_tetrahedra;

    if (get_orientability(manifold) != oriented_manifold){
      PyErr_SetString(PyExc_ValueError, 
		      "Triangulation must be oriented in SnapPeaC.get_gluing_equations");
      return NULL;
    }

    gluing_equations = PyList_New(0);      
    eqn = NEW_ARRAY(3 * T, int);

    /*
     *  Build edge equations.
     */

    for (   edge = manifold->edge_list_begin.next;
            edge != &manifold->edge_list_end;
            edge = edge->next)
      {
        for (i = 0; i < 3 * T; i++)
	  eqn[i] = 0;
	
        set_left_edge(edge, &ptet0);
        ptet = ptet0;
        do{
	  eqn[3*ptet.tet->index + edge3_between_faces[ptet.near_face][ptet.left_face]]++;
	  veer_left(&ptet);
        } while (same_positioned_tet(&ptet, &ptet0) == FALSE);
	
       
       /*  Now save to Python list */
	
       equation = int_array_to_PyList(eqn, 3*T);
       PyList_Append(gluing_equations, equation);
       Py_DECREF(equation);
		       
      }
    my_free(eqn);
    if ( PyErr_Occurred() ) return NULL;
    return gluing_equations;
}

/* computes the cusp equation the curve (merid)^m (long)^l 
 * in cusp cusp_num.  See get_gluing_equations for the return
 * convention.
 */

static PyObject *get_cusp_equation(PyObject *self, PyObject *args)
{
    long theTriangulation;
    Triangulation*  manifold = NULL;
    PyObject* equation;    

    int             *eqn,  i, coef[2], T, cusp_num, m, l;
    Tetrahedron     *tet;
    VertexIndex     v;
    Cusp *cusp;
    FaceIndex       f, ff;
    PeripheralCurve c;

    if (!PyArg_ParseTuple(args, "liii:get_gluing_equations",
			  &theTriangulation, &cusp_num, &m, &l))      {
        PyErr_SetString(PyExc_TypeError, 
			"Invalid arguments passed to SnapPeaC.get_cusp_equation\n"
			"           Expected ((long)triangulation), (int)  cusp_num, (int)  m, (int) l");
        return NULL;
    }

    manifold = (Triangulation*) theTriangulation;
    if (cusp_num < 0 || cusp_num >= get_num_cusps(manifold)){
      PyErr_SetString(PyExc_ValueError, "cusp_num not in valid range");
      return NULL;
    }

    if (get_orientability(manifold) != oriented_manifold){
      PyErr_SetString(PyExc_ValueError, 
		      "Triangulation must be oriented in SnapPeaC.get_cusp_equation");
      return NULL;
    }


    /* initialize variables */

    coef[0] = m;
    coef[1] = l;

    T = manifold -> num_tetrahedra;
    eqn = NEW_ARRAY(3 * T, int);
    for (i = 0; i < 3 * T; i++)
      eqn[i] = 0;

    /* find right cusp */

    cusp = manifold->cusp_list_begin.next;
    for (i = 0; i < cusp_num; i++ )
      cusp = cusp->next;
      

    /* compute equation */

    for (tet = manifold->tet_list_begin.next;
	 tet != &manifold->tet_list_end;
	 tet = tet->next){
      for (v = 0; v < 4; v++){
	if (tet->cusp[v] != cusp)
	  continue;

	for (f = 0; f < 4; f++){
	  if (f == v)
	    continue;

	  ff = remaining_face[v][f];

	  for (c = 0; c < 2; c++) /* c = M, L */
	    eqn[3*tet->index + edge3_between_faces[f][ff]]
	      += coef[c] *FLOW(tet->curve[c][right_handed][v][f], tet->curve[c][right_handed][v][ff]);
	}
      }
    }

    equation = int_array_to_PyList(eqn, 3*T);
    my_free(eqn);
    if ( PyErr_Occurred() ) return NULL;
    return equation;
}


static PyObject *wrap_same_triangulation(PyObject *self, PyObject *args)
{
    long theTriangulation1, theTriangulation2;
    Triangulation*  manifold1 = NULL;
    Triangulation*  manifold2 = NULL;

    if (!PyArg_ParseTuple(args, "ll:get_gluing_equations",
			  &theTriangulation1, &theTriangulation2))      {
        PyErr_SetString(PyExc_TypeError, 
			"Invalid arguments passed to SnapPeaC.wrap_same_triangulation\n"
			"           Expected ((long)triangulation), (long)triangulation)\n");
        return NULL;
    }
    manifold1   =(Triangulation*) theTriangulation1;
    manifold2  = (Triangulation*) theTriangulation2;
    if ( PyErr_Occurred() ) return NULL;
    return PyInt_FromLong(same_triangulation(manifold1, manifold2));
}


static PyObject *wrap_install_shortest_bases(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    if (!PyArg_ParseTuple(args, "l:install_shortest_bases", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_install_shortest_bases() in SnapPeaC.c received data of the wrong type.  Expected (long) triangulation");
        return NULL;
    }
        
    install_shortest_bases((Triangulation *)theTriangulation);
    if ( PyErr_Occurred() ) return NULL;
    return Py_None;
}


static PyObject *wrap_install_current_curve_bases(PyObject *self, PyObject *args)
{
    long            theTriangulation;
    if (!PyArg_ParseTuple(args, "l:install_current_curve_bases", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_install_current_curve_bases() in SnapPeaC.c received data of the wrong type.  Expected (long) triangulation");
        return NULL;
    }
        
    install_current_curve_bases((Triangulation *)theTriangulation);
    if ( PyErr_Occurred() ) return NULL;
    return Py_None;
}



static PyObject *wrap_sendable_to_triangulation(PyObject *self, PyObject *args)
{
    char* sendable;
    if (!PyArg_ParseTuple(args, "s:wrap_sendable_to_triangulation", &sendable))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_sendable_to_triangulation  in SnapPeaC.c received data of the wrong type.  Expected string");
        return NULL;
    }
        
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) sendable_to_triangulation(sendable));
}

static PyObject *wrap_hack_double_cover(PyObject *self, PyObject *args)
{
    long theTriangulation1;
    RepresentationList* reps;
    Triangulation*  manifold1 = NULL;
    Triangulation*  manifold2 = NULL;

    if (!PyArg_ParseTuple(args, "l:hack_double_cover",
			  &theTriangulation1))      {
        PyErr_SetString(PyExc_TypeError, "Invalid arguments passed to SnapPeaC.wrap_hack_double_cover.  Expected ((long)triangulation)\n");
        return NULL;
    }
    manifold1  = (Triangulation*) theTriangulation1;
    reps = find_representations(manifold1, 2, permutation_subgroup_Sn);
    manifold2 = construct_cover(manifold1, reps->list, reps->num_sheets);
    free_representation_list(reps);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) manifold2);
}

static PyObject *wrap_hack_create_all_covers(PyObject *self, PyObject *args)
{
    long theTriangulation1;
    int degree;
    RepresentationList* reps;
    RepresentationIntoSn* rep;
    Triangulation*  manifold1 = NULL;
    Triangulation*  manifold2 = NULL;
    PyObject *list, *cover;

    if (!PyArg_ParseTuple(args, "li:hack_double_cover",
			  &theTriangulation1, &degree))      {
        PyErr_SetString(PyExc_TypeError, "Invalid arguments passed to SnapPeaC.wrap_hack_create_all_covers.  Expected ((long)triangulation (int)degree)\n");
        return NULL;
    }
    manifold1  = (Triangulation*) theTriangulation1;
    reps = find_representations(manifold1, degree, permutation_subgroup_Sn);

    list = PyList_New(0);
    for (rep = reps->list; rep != NULL;  rep = rep->next){
      manifold2 = construct_cover(manifold1, rep, reps->num_sheets);
      cover = Py_BuildValue("l", (long int) manifold2);
      PyList_Append(list, cover);
    }
    
    free_representation_list(reps);
    
    if ( PyErr_Occurred() ) return NULL;
    return list;
}


static PyObject *wrap_choose_generators(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    int compute_corners, centroid_at_origin;

    if (!PyArg_ParseTuple(args, "lii", &theTriangulation, &compute_corners, &centroid_at_origin))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_choose_generators() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
        
    choose_generators((Triangulation *)theTriangulation, compute_corners, centroid_at_origin);
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}

static PyObject *wrap_hack_hide_hyperbolic_structure(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    Triangulation* manifold;
    if (!PyArg_ParseTuple(args, "l:hack_hide_hyperbolic_structure", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_hack_hide_hyperbolic_structure() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    manifold = (Triangulation*) theTriangulation;
    manifold->solution_type[complete] = not_attempted;    
    manifold->solution_type[filled] = not_attempted;
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}
 

static PyObject *wrap_find_complete_hyperbolic_structure(PyObject *self, PyObject *args)
{
    long    theTriangulation;
    if (!PyArg_ParseTuple(args, "l:find_complete_hyperbolic_structure", &theTriangulation))
    {
        PyErr_SetString(PyExc_TypeError, "wrap_find_complete_hyperbolic_structure() in SnapPeaC.c received data of the wrong type.");
        return NULL;
    }
    find_complete_hyperbolic_structure( (Triangulation* ) theTriangulation);
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("");
}


extern Triangulation*    triangulate_link_complement_from_file(char* file_name, char *path);


static PyObject *wrap_triangulate_link_complement_from_file(PyObject *self, PyObject *args)
{
  char *theName;
 Triangulation   *theTriangulation;
 
 if (!PyArg_ParseTuple(args, "s", &theName))
   {
     PyErr_SetString(PyExc_TypeError, "wrap_triangulate_link_complement_from_file() in SnapPeaC.c received data of the wrong type.");
        return NULL;
   }
 
 theTriangulation = triangulate_link_complement_from_file(theName, "");
    if (theTriangulation == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, "Could not create triangulation.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}


extern Triangulation*    fibered_manifold_associated_to_braid(int numStrands, int braidLength, int* word);

static PyObject *wrap_fibered_manifold_associated_to_braid(PyObject *self, PyObject *args)
{
  PyObject  *theBraidWord, *aPyInt;
  Triangulation   *theTriangulation;
  int numStrands, braidLength;
  int i;
  int *word;
 
  if (!PyArg_ParseTuple(args, "iO", &numStrands, &theBraidWord) || ! PyList_Check(theBraidWord))
   {
     PyErr_SetString(PyExc_TypeError, "wrap_fibered_manifold_associated_to_braid() in SnapPeaC.c received data of the wrong type.");
        return NULL;
   }
 
  braidLength = PyList_Size(theBraidWord);
  word = malloc(braidLength * sizeof(int));
  for (i = 0; i < braidLength; i++)
    {
      aPyInt = PyList_GetItem(theBraidWord, i);
      if (aPyInt == NULL )
        {
            PyErr_SetString(PyExc_RuntimeError, "Invalid braid word.");
            return NULL;
	}
        word[i] = (int)PyInt_AsLong(aPyInt);
    }
  
  theTriangulation = fibered_manifold_associated_to_braid(numStrands, braidLength, word);
  free(word);
    if (theTriangulation == NULL)
      {
        PyErr_SetString(PyExc_RuntimeError, "Could not create triangulation.");
        return NULL;
    }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theTriangulation);
}

/*

This function creates a cover of the given manifold, which is
specified by a list of permutations corresponding to each of the
generators of the simplified presentation of pi_1(M).  Each
permutation is stored simply as a list L where L[i] is the image of i
under the permutation.

*/

extern RepresentationIntoSn *initialize_new_representation(int num_original_generators, int n, int num_cusps);
extern RepresentationIntoSn *convert_candidateSn_to_original_generators(int **candidateSn, int n, int num_original_generators, int **original_generators, Triangulation *manifold, int **meridians, int **longitudes);
extern Boolean              candidateSn_is_valid(int **candidateSn, int n, int **group_relations, int num_relations);
extern Boolean              candidateSn_is_transitive(int **candidateSn, int num_generators, int n);
void                 free_representation(RepresentationIntoSn *representation, int num_generators, int num_cusps);


static PyObject *wrap_create_cover(PyObject *self, PyObject *args)
{
  PyObject  *thePermutations, *thePerm, *aPyInt;
  Triangulation   *theTriangulation, *theCover;
  int i,j, coverDegree;
  int numSimplifiedGens, numOriginalGens, num_simplified_relations;
  RepresentationIntoSn  *givenRepresentation, *repInOriginalGens;
  int **original_generators,   **simplified_group_relations,
    **meridians, **longitudes;
  GroupPresentation       *simplified_group;


  /* Check the arguments, and load the given permutation into givenRepresentation->image */

  if (!PyArg_ParseTuple(args, "lO", &theTriangulation, &thePermutations) || ! PyList_Check(thePermutations))
   {
     PyErr_SetString(PyExc_TypeError, "wrap_create_cover() in SnapPeaC.c received data of the wrong type.");
        return NULL;
   }
 
  simplified_group            = fundamental_group(theTriangulation, TRUE, TRUE, TRUE);
  numSimplifiedGens = PyList_Size(thePermutations);
  numOriginalGens = theTriangulation->num_generators;

  for (i = 0; i < numSimplifiedGens; i++)
    {
      thePerm = PyList_GetItem(thePermutations, i);
      if (!PyList_Check(thePerm))
	{
	  PyErr_SetString(PyExc_TypeError, "wrap_create_cover() in SnapPeaC.c received data of the wrong type.");
	  return NULL;
	}
    }
  
  coverDegree = PyList_Size(thePerm);
  givenRepresentation = initialize_new_representation(numOriginalGens, coverDegree, get_num_cusps(theTriangulation));
  
  for (i = 0; i < numSimplifiedGens; i++)
    {
      for (j = 0; j < coverDegree; j++)
	{
	  thePerm = PyList_GetItem(thePermutations, i);

	  aPyInt = PyList_GetItem(thePerm, j);
	  if (aPyInt == NULL )
	    {
	      PyErr_SetString(PyExc_RuntimeError, "Invalid permutation.");
	      return NULL;
	    }
	  givenRepresentation->image[i][j] = (int)PyInt_AsLong(aPyInt);
	}
    }

  if (numSimplifiedGens  != fg_get_num_generators(simplified_group))
    {
      PyErr_SetString(PyExc_RuntimeError, "Given permutation representation does not have the right number of generators");
      return NULL;
    }

  original_generators         = NEW_ARRAY(numOriginalGens, int *);
  for (i = 0; i < numOriginalGens; i++)
    original_generators[i] = fg_get_original_generator(simplified_group, i);
  
  num_simplified_relations    = fg_get_num_relations(simplified_group);
  simplified_group_relations  = NEW_ARRAY(num_simplified_relations, int *);
  for (i = 0; i < num_simplified_relations; i++)
    simplified_group_relations[i] = fg_get_relation(simplified_group, i);

  if ( (!candidateSn_is_valid(givenRepresentation->image, coverDegree, simplified_group_relations, num_simplified_relations))   ||  (!candidateSn_is_transitive(givenRepresentation->image, numSimplifiedGens, coverDegree)))
    {
      PyErr_SetString(PyExc_RuntimeError, "Given permutation representation does not actually give a transitive homorphism");
      return NULL;
    }
  
  meridians                   = NEW_ARRAY(theTriangulation->num_cusps, int *);
  longitudes                  = NEW_ARRAY(theTriangulation->num_cusps, int *);
  for (i = 0; i < theTriangulation->num_cusps; i++)
    {
        meridians [i] = fg_get_meridian (simplified_group, i);
        longitudes[i] = fg_get_longitude(simplified_group, i);
    }
    free_group_presentation(simplified_group);

    repInOriginalGens = convert_candidateSn_to_original_generators(givenRepresentation->image, coverDegree, numOriginalGens,  original_generators, theTriangulation, meridians, longitudes);
    
    /* Oh how much do I hate C?  Let me count the ways...  */

    for (i = 0; i < numSimplifiedGens; i++)
      fg_free_relation(simplified_group_relations[i]);
    my_free(simplified_group_relations);

    for (i = 0; i < numOriginalGens; i++)
      fg_free_relation(original_generators[i]);
    my_free(original_generators);

    for (i = 0; i <theTriangulation->num_cusps; i++)
    {
        fg_free_relation(meridians[i]);
        fg_free_relation(longitudes[i]);
    }
    my_free(meridians);
    my_free(longitudes);
    
    /* Finally, lets create the cover */
    
    theCover = construct_cover(theTriangulation, repInOriginalGens, coverDegree);
    
    free_representation(givenRepresentation, numSimplifiedGens, get_num_cusps(theTriangulation));
    free_representation(repInOriginalGens, numOriginalGens, get_num_cusps(theTriangulation));
  
  
    if (theCover == NULL)
      {
	PyErr_SetString(PyExc_RuntimeError, "Could not create cover.");
	return NULL;
      }
    
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("l", (long int) theCover);
}


static PyObject *wrap_get_cusp_neighborhood_info(PyObject *self, PyObject *args)
{
  long    theTriangulation, theIndex;
  double  cusp_volume;
  Complex meridian, longitude;
  CuspNeighborhoods* theCusp;
  PyObject* py_meridian, *py_longitude;
    
  if (!PyArg_ParseTuple(args, "li", &theTriangulation, &theIndex))
      {
        PyErr_SetString(PyExc_TypeError, "wrap_get_cusp_volume() in SnapPeaC.c received data of the wrong type.");
        return NULL;
      }
  theCusp = initialize_cusp_neighborhoods((Triangulation *) theTriangulation);
    if (theCusp == NULL)
      cusp_volume = -1;
    else
      {
	set_cusp_neighborhood_displacement(theCusp, theIndex, 1000.0);
	cusp_volume = get_cusp_neighborhood_cusp_volume(theCusp, theIndex);
	get_cusp_neighborhood_translations(theCusp, theIndex, &meridian, &longitude);
	py_meridian = PyComplex_FromDoubles(meridian.real,
					    meridian.imag);
	py_longitude = PyComplex_FromDoubles(longitude.real,
					     longitude.imag);

      }
    if ( PyErr_Occurred() ) return NULL;
    return Py_BuildValue("dOO", cusp_volume, py_meridian, py_longitude);
}


static PyObject *get_peripheral_curve_datum(PyObject *self, PyObject *args)
{
  Triangulation* theTriangulation;
  int i, tetIndex, whichCurve, vertexIndex, edgeIndex; 
  Tetrahedron* theTetrahedron;

    
  if (!PyArg_ParseTuple(args, "liiii", &theTriangulation, &tetIndex, &whichCurve, &vertexIndex, &edgeIndex))
      {
        PyErr_SetString(PyExc_TypeError, "get_peripheral_curve_datum() in SnapPeaC.c received data of the wrong type.");
        return NULL;
      }
  
  theTetrahedron = &theTriangulation->tet_list_begin;
  for(i = 0; i <= tetIndex; i++)
    theTetrahedron = theTetrahedron->next;
  return Py_BuildValue("ii", theTetrahedron->curve[whichCurve][right_handed][vertexIndex][edgeIndex], theTetrahedron->cusp[vertexIndex]->index);
}
